
//			CONTEXT command:


#ifndef CONTEXT_COMMAND_VERSION
#define CONTEXT_COMMAND_VERSION			1.0.1.1


#define MAX_CONTEXT_SENTENCES			16
#define CONTEXT_METAPHONE_LENGTH		8
#define CONTEXT_FLAG_REPORT				0x0001
#define CONTEXT_MAX_INDENT				20

#define WORD_TYPE_EXACT					0
#define WORD_TYPE_START					1
#define WORD_TYPE_METAPHONE				2
#define WORD_TYPE_SERIES				3
#define WORD_TYPE_SERIES_EXACT			4




const wchar_t ContextHelpText[] = L"Search for words in English text and display them in context.\n\
\n\
CONTEXT /C:n /CP:n /F:n /K:n /P /S /V /W:base /X:word /Y:word filename....\r\n\
\n\
\t/C:   sentences of context to display, before and after\n\
\t/CP:  interpret OEM text as code page n\n\
\t/F:   input text format:\n\
\t       0: best guess (default)\n\
\t       1: unformatted; honor all line breaks\n\
\t       2: prewrapped; skip single line breaks\n\
\t       3: unformatted; honor line breaks, no blank between grafs \n\
\t/K:   output columns for word-wrap\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\t/V    verbose; report counts after each file and at end\n\
\t/W:   search for forms of base word\n\
\t/X:   search for exact word\n\
\t/Y:   search for soundalike words\n\
\n\
You must specify at least one of /W: /X: or /Y:; you may have\n\
as many as you like.  Filenames may contain wildcards.\n";



struct SearchWord {
	SearchWord *Next;
	int Type;
	int Index;
	size_t StartPointer;
	unsigned int TotalCount, FileCount, Files;
	unsigned short int Flags;
	LPTSTR Plural;
	wchar_t Value;
};


struct FoundItem {
	FoundItem *Next;
	size_t StartOffset, EndOffset;
};


unsigned int ContextFilesProcessed		= 0;

wchar_t OutBuffer[SHORT_BUF_LEN]		= L"";
BYTE OutBufferColors[SHORT_BUF_LEN];




BOOL CTX_ChangeEnding( LPTSTR Word, LPTSTR OldEnding, LPTSTR NewEnding )
{
	if ( !Word || !OldEnding || !NewEnding )
		return FALSE;


	size_t BaseLen			= wcslen( Word );
	size_t OldEndingLen		= wcslen( OldEnding );


	if ( OldEndingLen > BaseLen )
		return FALSE;

	BaseLen -= OldEndingLen;

	if ( wcscmp( Word + BaseLen, OldEnding ) )
		return FALSE;

	Word[BaseLen] = '\0';
	wcscat_s( Word, SHORT_BUF_LEN, NewEnding );
	return TRUE;
}



BOOL MakeIrregPlural( LPTSTR inString, LPTSTR outString )
{
	if ( !inString || !outString )
		return FALSE;


	wchar_t temp[SHORT_BUF_LEN]			= L"";
	BOOL Changed						= FALSE;


	wcscpy_s( temp, SHORT_BUF_LEN, inString );
	_wcslwr_s( temp, SHORT_BUF_LEN );

	if ( CTX_ChangeEnding( temp, L"y", L"ies" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"is", L"es" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"ex", L"ices" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"s", L"ses" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"x", L"xes" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"o", L"oes" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"sh", L"shes" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"ch", L"ches" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"man", L"men" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"elf", L"elves" ) )
		Changed = TRUE;
	else if ( CTX_ChangeEnding( temp, L"ife", L"ives" ) )
		Changed = TRUE;

	if ( Changed )
		wcscpy_s( outString, SHORT_BUF_LEN, temp );

	return Changed;
}



int AddSearchWord( SearchWord **First, SearchWord **Last, int Type, LPTSTR Value )
{
	if ( !First || !Last )
		return -1;


	wchar_t Plural[SHORT_BUF_LEN]	= L"";
	
	size_t len						= wcslen( Value );
	size_t EntrySize				= sizeof( SearchWord ) + ( len * sizeof( wchar_t ) );


	if ( Type == WORD_TYPE_START ) {
		MakeIrregPlural( Value, Plural );
		if ( Plural[0] )
			EntrySize += ( ( wcslen( Plural ) + 1 ) * sizeof( wchar_t ) );
	}
	
	SearchWord *NewWord = NULL;
	NewWord = (SearchWord *) malloc( EntrySize );

	if ( !NewWord ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error allocating memory for search word\r\n" );
		return -2;
	}

	memset( NewWord, 0x00, EntrySize );

	NewWord->Type			= Type;

	wcscpy_s( &NewWord->Value, len + 1, Value );

	if ( Plural[0] ) {
		NewWord->Plural = &( NewWord->Value ) + len + 1;
		wcscpy_s( NewWord->Plural, wcslen( Plural ) + 1, Plural );
		//	wprintf( L"   * Also storing plural '%s' in entry:\n", NewWord->Plural );
	}

	if ( Type == WORD_TYPE_METAPHONE ) {
		wchar_t *sep1 = wcschr( &NewWord->Value, ':' );
		if ( sep1 )
			*sep1 = '\0';
	}

	if ( *Last != NULL )
		(*Last)->Next = NewWord;

	*Last = NewWord;

	if ( *First == NULL )
		*First = NewWord;
	
	return 0;
}



int ResetWordSeriesSearches( SearchWord *First )
{
	while ( First != NULL ) {
		if ( ( First->Type == WORD_TYPE_SERIES ) || ( First->Type == WORD_TYPE_SERIES_EXACT ) )
			First->Index = 0;

		First = First->Next;
	}

	return 0;
}



int ResetSearchFileCounts( SearchWord *First )
{
	while ( First != NULL ) {
		First->FileCount = 0;
		First->Flags = 0;

		First = First->Next;
	}

	return 0;
}



int InsertFoundItem( size_t Start, size_t End, FoundItem **FirstFind )
{
	FoundItem *NewItem = NULL;
	NewItem = (FoundItem *) malloc( sizeof( FoundItem ) );

	if ( !NewItem )
		return -2;

	NewItem->Next			= NULL;
	NewItem->StartOffset	= Start;
	NewItem->EndOffset		= End;

	if ( *FirstFind == NULL ) {
		*FirstFind = NewItem;
		return 0;
	}
	else if ( Start < (*FirstFind)->StartOffset ) {
		NewItem->Next = *FirstFind;
		*FirstFind = NewItem;
		return 0;
	}

	FoundItem *ThisItem = *FirstFind, *PrevItem = NULL, *NextItem = (*FirstFind)->Next;

	while ( NextItem ) {
		if ( Start < NextItem->StartOffset ) {
			NewItem->Next = NextItem;
			ThisItem->Next = NewItem;
			return 0;
		}

		ThisItem = NextItem;
		NextItem = ThisItem->Next;
	}

	ThisItem->Next = NewItem;
	return 0;
}



int RemoveFirstFoundItem( FoundItem **FirstFind )
{
	if ( !FirstFind )
		return -1;


	FoundItem *This		= *FirstFind;
	FoundItem *Next		= This ? This->Next : NULL;

	free( This );
	*FirstFind = Next;
	return 0;
}



int CTX_Dump( HANDLE ConsoleHandle, BYTE Normal, BYTE Highlight )
{
	wchar_t Temp[SHORT_BUF_LEN + 1] = L"";
	wcscpy_s( Temp, SHORT_BUF_LEN + 1, OutBuffer );
	TrimRight( Temp );


	if ( Normal == 0 ) {
		QPuts( Temp );
		return CL_EndLine();
	}


	BYTE sv = 0;

	for ( int i = 0; Temp[i]; i++ ) {
		if ( OutBufferColors[i] != sv ) {
			sv = OutBufferColors[i];
			SetConsoleTextAttribute( ConsoleHandle, sv );
		}
		_putwch( OutBuffer[i] );
	}

	if ( sv != Normal )
		SetConsoleTextAttribute( ConsoleHandle, Normal );

	return CL_EndLine();
}



int RemoveChars( int n )
{
	if ( n < 1 )
		return 0;

	if ( n > (int) wcslen( OutBuffer ) )
		return 0;

	int i = n;
	wchar_t ch = SPACE;
	BYTE sv = 0x00;

	while ( OutBuffer[i] == SPACE )
		i++;

	for ( int j = 0; ch != 0x00; j++ ) {
		sv = OutBufferColors[i + j];
		OutBufferColors[j] = sv;
		ch = OutBuffer[i + j];
		OutBuffer[j] = ch;
	}

	return i;
}



int CTX_DumpBuffer( _MMFile *MMFile, size_t StartIndex, size_t EndIndex, int Width, HANDLE ConsoleHandle, BYTE Normal, BYTE Highlight, FoundItem **FirstFind )
{
	size_t SaveIndex			= MMFile->Index;

	OutBuffer[0]				= '\0';
	int OutCount				= 0;
	size_t OldIndex				= StartIndex;
	size_t TestIndex			= NULL;
	int ch						= 0;
	int lastch					= 0;
	int sv						= 0;
	int sv2						= 0x00;
	int LastSpace				= -1;
	int LastHyphen				= -1;
	wchar_t LastHyphenChar		= 0x00;
	int HyphenMin = 3;
	int crcount = 0, lfcount = 0;
	int i = 0;
	BOOL EatSpaces = FALSE;
	int TabSize = 4;
	int CharSize = 1;


	if ( ( MMFile->FileEncoding == FILE_UTF16LE ) || ( MMFile->FileEncoding == FILE_UTF16BE ) )
		CharSize = 2;
	else if ( ( MMFile->FileFormat == FILE_UTF32LE ) || ( MMFile->FileEncoding == FILE_UTF32BE ) )
		CharSize = 4;

	MMFile->Index		= StartIndex;


	while ( ( MMPeekAhead( MMFile, 0 ) > 0 ) && iswcntrl( MMPeekAhead( MMFile, 0 ) ) )
		MMGetChar( MMFile );										//  skip over any control characters at the start of the region

	while ( MMFile->Index < ( EndIndex - CharSize ) ) {				//  then backtrack over any control characters
		ch = MMPrevChar( MMFile, EndIndex, &TestIndex );			//	at the end of the region

		if ( ( ch >= 0 ) && iswcntrl( ch ) )
			EndIndex = TestIndex;
		else
			break;
	}

	ch = SPACE;

	do {
		OldIndex = MMFile->Index;
		ch = MMGetChar( MMFile );

		if ( ch == 0x00 )
			continue;

		if ( EatSpaces && ( ch == SPACE ) )
			continue;

		if ( ch != SPACE )
			EatSpaces = FALSE;


		//  If the character is a soft hyphen, deal with it:

		if ( ( ch == SHYPHEN ) && ( OutCount > 0 ) && ( OutCount > ( Width - 1 ) ) ) {
			if ( iswalpha( OutBuffer[OutCount - 1] ) && iswalnum( MMPeekAhead( MMFile, 0 ) ) ) {
				BOOL Ok = TRUE;
				int LookAhead = 0;

				for ( int j = 0; j < HyphenMin; j++ ) {		// count chars to move to next line
					LookAhead = MMPeekAhead( MMFile, j );
					if ( ( LookAhead <= 0 ) || iswspace( LookAhead ) || _MMIsEOL( LookAhead ) )
						Ok = FALSE;
				}

				if ( Ok ) {						// if enough characters would move to the next line,
					LastHyphen = OutCount;		// note the soft hyphen's location
					LastHyphenChar = ch;		// and value
				}
			}
		}


		// Expand tabs:

		if ( ch == TAB ) {
			int k = TabSize - ( OutCount % TabSize );
			for ( i = 0; i < k; i++ ) {
				if ( OutCount < ( MAX_ARG_LEN - 1 ) ) {
					LastSpace = OutCount;
					OutBufferColors[OutCount] = Normal;
					OutBuffer[OutCount++] = SPACE;
					OutBuffer[OutCount] = '\0';
				}
			}
			continue;
		}


		if ( ch == SHYPHEN )			// discard soft hyphens
			continue;


		//  Handle line terminators:
		if ( _MMIsEOL( ch ) ) {
			if ( MMPeekAhead( MMFile, 0 ) == _MM2PartEOL( ch ) )
				MMGetChar( MMFile );

			if ( ( MMFile->FileFormat == FMT_PREWRAPPED ) && !_MMIsEOL( MMPeekAhead( MMFile, 0 ) ) ) {
				ch = SPACE;
				EatSpaces = TRUE;

				if ( OutCount ) {
					if ( OutBuffer[OutCount - 1] == SPACE )
						continue;
				}
				else
					continue;

				if ( OutCount > 1 )
					if ( iswalpha( OutBuffer[OutCount - 2] ) && ( IsHyphen( OutBuffer[OutCount - 1] ) ) ) {
						i = 0;
						while ( ( i < MAX_INDENT ) && ( ( MMPeekAhead( MMFile, i ) == SPACE ) || ( MMPeekAhead( MMFile, i ) == TAB ) ) )
							i++;

						continue;
					}
			}
		}


		if ( _MMIsEOL( ch ) ) {													// end of paragraph:
			if ( CTX_Dump( ConsoleHandle, Normal, Highlight ) == USER_ABORT )	// display contents of buffer
				return USER_ABORT;

			if ( ( MMFile->FileFormat == FMT_UNFORMATTED ) && ( OutCount > 0 ) ) {
				if ( CL_EndLine() == USER_ABORT )			// add a blank line
					return USER_ABORT;
			}
			else if ( MMFile->FileFormat == FMT_PREWRAPPED ) {
				if ( CL_EndLine() == USER_ABORT )			// add a blank line
					return USER_ABORT;
			}

			OutCount = 0;									// empty the output buffer
			OutBuffer[OutCount] = '\0';
			continue;
		}

		if ( ( ch != NBSP ) && iswspace( ch ) ) {		// space in input text:
			if ( !StuffNotToBreakBefore( MMPeekAhead( MMFile, 0 ), MMPeekAhead( MMFile, 1 ) ) ) {
				LastSpace = OutCount;					// note its location
				LastHyphen = -1;						// and discard any pending hyphen break
				LastHyphenChar = 0x00;
			}
		}


		//  If the character is a hyphen, deal with it:

		if ( ( OutCount > 0 ) && ( OutCount < ( Width - 1 ) ) && ( IsHyphen( ch ) ) )
			if ( iswalpha( OutBuffer[OutCount - 1] ) && iswalnum( MMPeekAhead( MMFile, 0 ) ) ) {
				BOOL Ok = TRUE;
				int LookAhead = 0x00;

				for ( int j = 0; j < HyphenMin; j++ ) {		// count chars to move to next line
					LookAhead = MMPeekAhead( MMFile, j );
					if ( ( LookAhead <= 0 ) || iswspace( LookAhead ) || _MMIsEOL( LookAhead ) )
						Ok = FALSE;
				}

				if ( Ok ) {						// if enough characters would move to the next line,
					LastHyphen = OutCount;		// note the hyphen's location
					LastHyphenChar = ch;		// and value
				}
			}


		//  Add character to the output buffer:

		if ( OutCount < ( SHORT_BUF_LEN - 1 ) ) {
			OutBufferColors[OutCount] = Normal;

			if ( *FirstFind ) {
				if ( OldIndex >= (*FirstFind)->EndOffset ) {
					RemoveFirstFoundItem( FirstFind );		//	RemoveItemFromList( &FirstFind );

					while ( *FirstFind ) {
						if ( OldIndex >= (*FirstFind)->EndOffset ) {
							RemoveFirstFoundItem( FirstFind );		//	RemoveItemFromList( &FirstFind );
						}
						else {
							if ( OldIndex >= (*FirstFind)->StartOffset )
								OutBufferColors[OutCount] = Highlight;

							break;
						}
					}
				}
				else if ( OldIndex >= (*FirstFind)->StartOffset )
					OutBufferColors[OutCount] = Highlight;
			}

			OutBuffer[OutCount++] = ch;
			OutBuffer[OutCount] = '\0';
		}


		if ( OutCount >= Width ) {		// if the output buffer is full ....

			if ( ( LastHyphen >= ( Width * 2 / 5 ) ) && ( LastHyphen > ( LastSpace + HyphenMin ) ) ) {
				sv = OutBuffer[LastHyphen + 1];			// break at last hyphen
				sv2 = OutBuffer[LastHyphen];

				OutBuffer[LastHyphen + 1] = 0x00;
				if ( LastHyphenChar == SHYPHEN )
					OutBuffer[LastHyphen] = HYPHEN;

				if ( CTX_Dump( ConsoleHandle, Normal, Highlight ) == USER_ABORT )
					return USER_ABORT;

				OutBuffer[LastHyphen + 1] = sv;
				OutBuffer[LastHyphen] = sv2;

				if ( LastHyphenChar == SHYPHEN )
					OutCount -= RemoveChars( LastHyphen );
				else
					OutCount -= RemoveChars( LastHyphen + 1 );
			}
			else if ( LastSpace >= ( Width * 2 / 5 ) ) {
				sv = OutBuffer[LastSpace];				// break at last space
				OutBuffer[LastSpace] = 0x00;

				if ( CTX_Dump( ConsoleHandle, Normal, Highlight ) == USER_ABORT )
					return USER_ABORT;

				OutBuffer[LastSpace] = sv;
				OutCount -= RemoveChars( LastSpace + 1 );

				if ( OutCount == 0 )
					EatSpaces = TRUE;
			}
			else {										// break in mid-word without apology
				sv = OutBuffer[Width - 2];
				sv2 = OutBuffer[Width - 1];
				OutBuffer[Width - 2] = HYPHEN;
				OutBuffer[Width - 1] = 0x00;

				if ( CTX_Dump( ConsoleHandle, Normal, Highlight ) == USER_ABORT )
					return USER_ABORT;

				OutBuffer[Width - 2] = sv;
				OutBuffer[Width - 1] = sv2;
				OutCount -= RemoveChars( Width - 2 );
			}

			LastSpace = LastHyphen = -1;
			LastHyphenChar = 0x00;
		}

	} while ( MMFile->Index < EndIndex );

	if ( OutCount )
		if ( CTX_Dump( ConsoleHandle, Normal, Highlight ) == USER_ABORT )
			return USER_ABORT;

	MMFile->Index = SaveIndex;
	return 0;
}



int CTX_DividerLine( int Width )
{
	int n = ( Width - 11 ) / 2;

	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	for ( int i = 0; i < n; i++ )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), SPACE );

	QPuts( L"*    *    *" );

	if ( CL_EndLine( 2 ) == USER_ABORT )
		return USER_ABORT;

	return 0;
}



int ForceSentenceEnd( _MMFile *MMFile, size_t Pointer, wchar_t Char )
{
	int ch = '\0';
	unsigned int MaxCRs = 0, MaxLFs = 0;

	if ( MMFile->FileFormat == FMT_PREWRAPPED )
		MaxCRs = MaxLFs = 1;

	for ( unsigned int i = 0; i < 24; i++ ) {
		ch = MMPeekAhead( MMFile, i );

		if ( ch < 0 )
			return 1;
		else if ( iswalpha( ch ) ) {
			if ( islower( ch ) )
				return 0;
			else
				return 1;
		}
		else if ( IsTerminalPunct( ch ) )
			return 0;
		else if ( ( Char == ELLIPSIS ) && ( ch == '.' ) )
			return 1;
		else if ( ( ( Char == '.' ) || ( Char == FWPERIOD ) ) && ( ch == ELLIPSIS ) )
			return 1;
		else if ( ch == CR ) {
			if ( MaxCRs == 0 )
				return 1;
			else
				MaxCRs--;
		}
		else if ( ch == LF ) {
			if ( MaxLFs == 0 )
				return 1;
			else
				MaxLFs--;
		}
	}

	return 0;
}



int ContextShowFinalReport( SearchWord *First )
{
	BOOL Plural = FALSE;


	if ( HLprintf( L"\n\x12\rAll %u files:\n\n", ContextFilesProcessed ) == USER_ABORT )
		return USER_ABORT;

	while ( First ) {

		Plural = ( First->TotalCount != 1 );

		if ( First->TotalCount == 0 )
			QPuts( L"  Found no " );
		else
			Printf( L"  Found %u ", First->TotalCount );

		switch ( First->Type ) {

			case WORD_TYPE_EXACT :
				Printf( L"exact match%s to", Plural ? L"es" : L"" );
				break;

			case WORD_TYPE_START :
				Printf( L"word%s that might be a form of", Plural ? L"s" : L"" );
				break;

			case WORD_TYPE_METAPHONE :
				Printf( L"word%s sounding like", Plural ? L"s" : L"" );
				break;

			case WORD_TYPE_SERIES :
				QPuts( L"word series matching" );
				break;

			case WORD_TYPE_SERIES_EXACT :
				QPuts( L"exact word series matching" );
				break;
		}

		if ( First->Type == WORD_TYPE_METAPHONE )
			Printf( L" \"%s\" (%s)", &First->Value + wcslen( &First->Value ) + 1, &First->Value );
		else
			Printf( L" \"%s\"", &First->Value );

		if ( First->TotalCount )
			Printf( L" in %u file%s.", First->Files, ( First->Files != 1 ) ? L"s" : L"" );
		else
			QPuts( L"." );

		if ( CL_EndLine() == USER_ABORT )
			return USER_ABORT;

		First = First->Next;
	}

	return 0;
}



int ContextShowFileReport( SearchWord *First, LPCTSTR Filename )
{
	BOOL Plural = FALSE;


	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	if ( !wcscmp( Filename, ClipFilename ) )
		QPuts( L"  End of <clipboard>." );
	else if ( !wcscmp( Filename, StdinFilename ) )
		QPuts( L"  End of <stdin>." );
	else
		Printf( L"  End of file \"%s\".", Filename );

	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;


	while ( First ) {

		Plural = ( First->FileCount != 1 );

		if ( First->FileCount == 0 )
			QPuts( L"  Found no " );
		else
			Printf( L"  Found %u ", First->FileCount );

		switch ( First->Type ) {

			case WORD_TYPE_EXACT :
				Printf( L"exact match%s to", Plural ? L"es" : L"" );
				break;

			case WORD_TYPE_START :
				Printf( L"word%s that might be a form of", Plural ? L"s" : L"" );
				break;

			case WORD_TYPE_METAPHONE :
				Printf( L"word%s sounding like", Plural ? L"s" : L"" );
				break;

			case WORD_TYPE_SERIES :
				QPuts( L"word series matching" );
				break;

			case WORD_TYPE_SERIES_EXACT :
				QPuts( L"exact word series matching" );
				break;
		}

		if ( First->Type == WORD_TYPE_METAPHONE )
			Printf( L" \"%s\" (%s).", &First->Value + wcslen( &First->Value ) + 1, &First->Value );
		else
			Printf( L" \"%s\".", &First->Value );

		if ( CL_EndLine() == USER_ABORT )
			return USER_ABORT;

		First = First->Next;
	}

	return 0;
}



int FormOfWord( LPTSTR BaseWord, LPTSTR BaseWordPlural, LPTSTR WordToTest )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	int cmp					= (int) wcslen( BaseWord );

	wchar_t LastLetter		= towlower( BaseWord[cmp - 1] );
	BOOL MatchStart			= FALSE;
	int ExtraVowels			= 0;


	MatchStart = ( _wcsnicmp( BaseWord, WordToTest, cmp ) == 0 );

	if ( !MatchStart ) {
		if ( LastLetter == 'y' ) {
			if ( towlower( WordToTest[cmp - 1] ) == 'i' )
				MatchStart = ( _wcsnicmp( BaseWord, WordToTest, cmp - 1 ) == 0 );
		}
		else if ( LastLetter == 'e' ) {
			if ( towlower( WordToTest[cmp - 1] ) == 'y' )
				if ( WordToTest[cmp] == '\0' )
					MatchStart = ( _wcsnicmp( BaseWord, WordToTest, cmp - 1 ) == 0 );
		}
	}

	if ( MatchStart ) {
		for ( int i = cmp; WordToTest[i]; i++ ) {
			wchar_t ch = towlower( WordToTest[i] );
			if ( ( ch == 'a' ) || ( ch == 'e' ) || ( ch == 'i' ) || ( ch == 'o' ) || ( ch == 'u' ) )
				ExtraVowels++;
		}

		if ( ExtraVowels < 2 )
			return 1;
	}

	if ( !BaseWordPlural ) {
		MakeIrregPlural( BaseWord, temp );
		BaseWordPlural = temp;
	}

	if ( BaseWordPlural[0] )
		if ( !_wcsicmp( WordToTest, BaseWordPlural ) )
			return 1;

	return 0;
}



int ContainsFormOfWord( LPTSTR BaseWord, LPTSTR BaseWordPlural, LPTSTR WordToTest )
{
	if ( !BaseWord || !WordToTest )
		return 0;
	else if ( ( BaseWord[0] == '\0' ) || ( WordToTest[0] == '\0' ) )
		return 0;

	if ( FormOfWord( BaseWord, BaseWordPlural, WordToTest ) )
		return 1;


	wchar_t Temp[SHORT_BUF_LEN] = L"";
	wcscpy_s( Temp, SHORT_BUF_LEN, WordToTest );

	wchar_t *SubWordStart = Temp, *SubWordEnd = wcschr( SubWordStart, HYPHEN );

	if ( !SubWordEnd )
		return 0;

	do {
		if ( SubWordEnd )
			*SubWordEnd = '\0';

		if ( FormOfWord( BaseWord, BaseWordPlural, SubWordStart ) )
			return 1;

		if ( !SubWordEnd )
			break;

		SubWordStart = SubWordEnd + 1;

		if ( *SubWordStart == '\0' )
			break;

		SubWordEnd = wcschr( SubWordStart, HYPHEN );

	} while ( SubWordStart != NULL );
	
	return 0;
}



int WordSeriesMatch( LPTSTR WordSeries, LPTSTR WordToTest, SearchWord *Entry, size_t *SeriesStartPointer )
{
	if ( !WordSeries || !WordToTest )
		return 0;
	else if ( ( WordSeries[0] == '\0' ) || ( WordToTest[0] == '\0' ) )
		return 0;

	int i = 0;
	wchar_t *WordStart = &Entry->Value;
	wchar_t *WordEnd = wcschr( WordStart, SPACE );

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	wchar_t ch = SPACE;
	int j = 0;
	BOOL Match = FALSE;

	while ( ( Entry->Index ) > i ) {
		if ( WordEnd == NULL )
			return 0;

		WordStart = WordEnd + 1;
		if ( *WordStart == '\0' )
			return 0;

		WordEnd = wcschr( WordStart, SPACE );
		i++;
	}

	for ( int k = 0; ( j < SHORT_BUF_LEN ) && ch; k++ ) {
		ch = WordStart[k];
		if ( ch == SPACE )
			ch = '\0';

		if ( j < SHORT_BUF_LEN )
			Temp[j++] = ch;
	}
	Temp[SHORT_BUF_LEN - 1] = '\0';

	if ( Entry->Type == WORD_TYPE_SERIES_EXACT )
		Match = ( _wcsicmp( Temp, WordToTest ) == 0 );
	else
		Match = FormOfWord( Temp, NULL, WordToTest );

	if ( Match ) {
		if ( Entry->Index == 0 )
			Entry->StartPointer = *SeriesStartPointer;

		if ( WordEnd == NULL ) {
			Entry->Index = 0;
			*SeriesStartPointer = Entry->StartPointer;
			return 1;
		}
		else 
			Entry->Index = Entry->Index + 1;
	}
	else
		Entry->Index = 0;
	
	return 0;
}



int CompareWords( LPTSTR inString, SearchWord *FirstWord, size_t *SeriesStartPointer )
{
	if ( !inString || !FirstWord )
		return 0;

	if ( inString[0] == '\0' )
		return 0;

	wchar_t MetaphoneBuffer[20] = L"";
	int rv = 0;
	BOOL Match = FALSE;

	while ( FirstWord != NULL ) {
		Match = FALSE;

		switch ( FirstWord->Type ) {
			case WORD_TYPE_EXACT :
				if ( _wcsicmp( inString, &FirstWord->Value ) == 0 )
					Match = TRUE;

				break;

			case WORD_TYPE_START :
				if ( ContainsFormOfWord( &(FirstWord->Value), FirstWord->Plural, inString ) )
					Match = TRUE;

				break;

			case WORD_TYPE_METAPHONE :
				Metaphone( inString, MetaphoneBuffer, CONTEXT_METAPHONE_LENGTH, 0 );

				if ( wcscmp( MetaphoneBuffer, &FirstWord->Value ) == 0 )
					Match = TRUE;

				break;

			case WORD_TYPE_SERIES :
			case WORD_TYPE_SERIES_EXACT :
				if ( WordSeriesMatch( &FirstWord->Value, inString, FirstWord, SeriesStartPointer ) )
					Match = TRUE;

				break;
		}

		if ( Match ) {
			FirstWord->TotalCount++;
			FirstWord->FileCount++;

			if ( ( FirstWord->Flags & 0x01 ) == 0 ) {
				FirstWord->Flags |= 0x01;
				FirstWord->Files++;
			}

			rv = 1;
		}

		FirstWord = FirstWord->Next;
	}

	return rv;
}



struct FindWordsArgs {
	SearchWord *FirstWord;
	unsigned int UserFormat;
	int SentencesBefore;
	int SentencesAfter;
	int Width;
	BYTE NormalColor;
	BYTE HighlightColor;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



int FindWords( LPCTSTR Filename, int Flags, void *ContextArgs )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_MMFile MMFile;
	FoundItem *FirstFind		= NULL;
	FindWordsArgs *MoreArgs		= (FindWordsArgs*) ContextArgs;
	SearchWord *FirstWord		= MoreArgs->FirstWord;
	unsigned int UserFormat		= MoreArgs->UserFormat;
	int SentencesBefore			= MoreArgs->SentencesBefore;
	int SentencesAfter			= MoreArgs->SentencesAfter;
	int Width					= MoreArgs->Width;
	BOOL ShowReport				= Flags & CONTEXT_FLAG_REPORT;

	HANDLE ConHandle			= NULL;


	int rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	int ContextCountdown = 0;
	size_t OldIndex = 0, WordStartIndex = 0;
	size_t SentenceStartIndex[MAX_CONTEXT_SENTENCES];
	size_t DumpStartIndex = 0, DumpEndIndex = 0;

	BOOL FoundWords = FALSE;

	int ch = SPACE, lastch = 0x00, ch2 = 0x00;
	wchar_t WordBuf[SHORT_BUF_LEN] = L"";
	int WordLen = 0;
	int WordPeriods = 0, WordAlnum = 0, WordVowels = 0;
	BOOL InSingleQuotes = FALSE;
	wchar_t AposWord[SHORT_BUF_LEN] = L"";

	BOOL BlankLine = TRUE;
	BOOL FileIsPrewrapped = FALSE;
	int BlankLinesCount		= 0;
	int BlankLinesReset		= 2;


	for ( int i = 0; i < MAX_CONTEXT_SENTENCES; i++ )
		SentenceStartIndex[i] = MMFile.Index;
	
	if ( UserFormat )
		MMFile.FileFormat = UserFormat;
	else
		GuessFileFormat( &MMFile );

	FileIsPrewrapped = ( MMFile.FileFormat == FMT_PREWRAPPED );


	int Slack = Width * 4 / 5;
	if ( ( MMFile.FileEncoding == FILE_UTF16LE ) || ( MMFile.FileEncoding == FILE_UTF16BE ) )
		Slack = Slack * 2;


	if ( MoreArgs->NormalColor )
		ConHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, NULL, NULL );


	//  Display the filename:

	if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
		goto UserAbort;


	//  Reset search file counts:

	ResetSearchFileCounts( FirstWord );


	//  Loop through the contents of the file:

	while ( MMFile.Index < MMFile.FileSize ) {
		lastch = ch;
		OldIndex = MMFile.Index;
		ch = MMGetChar( &MMFile );

		if ( !BlankLine && ( ch < 0 ) )
			ch = CR;

		if ( ch < 0 )
			break;					// abort on error or eof
		else if ( ch == '\0' )
			continue;				// ignore any nulls in input text

		if ( iswgraph( ch ) )
			BlankLine = FALSE;

		if ( ( ch == SHYPHEN ) && iswalnum( lastch ) )
			if ( iswalnum( MMPeekAhead( &MMFile, 0 ) ) )
				continue;			// ignore soft hyphens between alphanumerics

		if ( ( ch == HYPHEN2 ) || ( ch == SHYPHEN ) )
			ch = HYPHEN;

		if ( ( ch == LSQUOTE ) || ( ch == RSQUOTE ) || ( ch == APOSTROPHE ) ) {
			ch = APOSTROPHE;
			wchar_t nextch = MMPeekAhead( &MMFile, 0 );

			if ( iswgraph( lastch ) && ( lastch != LDQUOTE ) && ( lastch != QUOTE ) ) {
				ch = RSQUOTE;

				if ( iswalnum( lastch ) && ( iswalpha( nextch ) ) )
					ch = APOSTROPHE;

				if ( !InSingleQuotes && iswalpha( lastch ) && ( ch == RSQUOTE ) )
					ch = APOSTROPHE;

				if ( ch == RSQUOTE )
					InSingleQuotes = FALSE;
			}
			else {
				ch = LSQUOTE;

				if ( _MMIsEOL( nextch ) )
					ch = RSQUOTE;
			}

			if ( ( iswalnum( lastch ) == 0 ) && ( iswalpha( nextch ) ) ) {
				AposWord[0] = APOSTROPHE;
				int peek = 0;
				for ( int j = 0; j < 15; j++ ) {
					peek = MMPeekAhead( &MMFile, j );
					AposWord[j + 1] = ( peek < SPACE ) ? SPACE : peek;
				}

				AposWord[16] = 0x00;
				if ( IsWordBeginningWithApostrophe( AposWord ) )
					ch = APOSTROPHE;
			}

			if ( ch == LSQUOTE )
				InSingleQuotes = TRUE;
		}

		if ( !InSingleQuotes && ( ch == '`' ) && ( iswgraph( lastch ) == 0 ) ) {
			if ( iswalnum( MMPeekAhead( &MMFile, 0 ) ) )
				InSingleQuotes = TRUE;
			else if ( MMPeekAhead( &MMFile, 0 ) == '`' )
				if ( iswalnum( MMPeekAhead( &MMFile, 1 ) ) )
					InSingleQuotes = TRUE;
		}


		if ( ch == '.' ) {
			if ( WordLen ) {
				if ( ( WordVowels > 0 ) && ( WordPeriods == 0 ) )
					if ( iswalpha( MMPeekAhead( &MMFile, 0 ) ) == 0 )
						ch = FWPERIOD;

				if ( ( ch == FWPERIOD ) && !iswalpha( MMPeekAhead( &MMFile, 0 ) ) ) {
					if ( _wcsicmp( WordBuf, L"etc" ) == 0 )
						ch = '.';
					else if ( _wcsicmp( WordBuf, L"viz" ) == 0 )
						ch = '.';
				}

				if ( ch == '.' )
					WordPeriods++;
			}
			else
				ch = FWPERIOD;
		}

		if ( WordLen && ( ch == HYPHEN ) )
			if ( iswpunct( MMPeekAhead( &MMFile, 0 ) ) )
				ch = HYPHEN2;

		if ( _MMIsEOL( ch ) && WordLen && FileIsPrewrapped && ( lastch == HYPHEN ) ) {
			int i = 0;

			if ( MMPeekAhead( &MMFile, i ) == _MM2PartEOL( ch ) )
				i++;

			while ( ( i < 20 ) && IsSpaceOrTab( MMPeekAhead( &MMFile, i ) ) )
				i++;

			if ( iswalnum( MMPeekAhead( &MMFile, i ) ) ) {

				for ( int j = 0; j < i; j++ )
					MMGetChar( &MMFile );

				ch = HYPHEN;

				if ( WordLen > 1 ) {
					WordLen--;
					WordBuf[WordLen] = 0x00;
					ch = WordBuf[WordLen - 1];
				}

				continue;
			}
		}

		if ( iswalnum( ch ) || ( ch == APOSTROPHE ) || ( ( WordLen > 0 ) && ( ( ch == '.' ) || ( ch == HYPHEN ) ) ) ) {
			if ( WordLen == 0 )
				WordStartIndex = OldIndex;

			if ( WordLen < ( SHORT_BUF_LEN - 1 ) ) {
				WordBuf[WordLen++] = StripAccent( ch );
				WordBuf[WordLen] = 0x00;

				if ( iswalnum ( ch ) )
					WordAlnum++;

				if ( IsVowel( ch ) )
					WordVowels++;
			}
		}
		else if ( WordLen ) {
			if ( WordAlnum ) {
				if ( WordLen > 1 )
					if ( WordBuf[WordLen - 1] == HYPHEN )
						WordBuf[--WordLen] = 0x00;

				//	wprintf( L"   * Found word %s\n", WordBuf );

				if ( CompareWords( WordBuf, FirstWord, &WordStartIndex ) ) {
					FoundWords = TRUE;

					if ( MoreArgs->HighlightColor ) {
						InsertFoundItem( WordStartIndex, OldIndex, &FirstFind );
					}

					if ( DumpEndIndex && ( SentenceStartIndex[SentencesBefore] > ( DumpEndIndex + Slack ) ) ) {
						if ( CTX_DumpBuffer( &MMFile, DumpStartIndex, DumpEndIndex, Width, ConHandle, MoreArgs->NormalColor, MoreArgs->HighlightColor, &FirstFind ) == USER_ABORT )
							goto UserAbort;

						DumpStartIndex = SentenceStartIndex[SentencesBefore];
						if ( CTX_DividerLine( Width ) == USER_ABORT )
							goto UserAbort;
					}
					else if ( DumpEndIndex == 0 )
						DumpStartIndex = SentenceStartIndex[SentencesBefore];

					DumpEndIndex = MMFile.Index;
					ContextCountdown = SentencesAfter + 1;
				}

				if ( iswalpha( WordBuf[0] ) && iswlower( WordBuf[0] ) && ( WordBuf[WordLen - 1] == '.' ) )
					if ( ForceSentenceEnd( &MMFile, MMFile.Index, '.' ) )
						ch = FWPERIOD;
			}

			WordLen = 0;
			WordBuf[0] = '\0';
			WordAlnum = WordVowels = WordPeriods = 0;
		}

		if ( ch == ELLIPSIS ) {
			if ( ForceSentenceEnd( &MMFile, MMFile.Index, ch ) )
				ch = FWPERIOD;
		}
		else if ( ch == FWPERIOD )
			if ( ForceSentenceEnd( &MMFile, MMFile.Index, '.' ) == 0 )
				ch = '.';

		if ( IsTerminalPunct( ch ) || ( ch == FWPERIOD ) ) {
			ch2 = MMPeekAhead( &MMFile, 0 );

			if ( ch == FWPERIOD ) {
				while ( ( ( ch2 == SPACE ) || ( ch2 == NBSP ) ) && ( IsTerminalPunct( MMPeekAhead( &MMFile, 1 ) ) || ( MMPeekAhead( &MMFile, 1 ) == '.' ) ) ) {
					MMGetChar( &MMFile );
					ch = MMGetChar( &MMFile );
					ch2 = MMPeekAhead( &MMFile, 0 );
				}
			}

			while ( IsTerminalPunct( ch2 ) || ( ch2 == '.' ) ) {
				ch = MMGetChar( &MMFile );
				ch2 = MMPeekAhead( &MMFile, 0 );
			}

			while ( ( ch2 == QUOTE ) || ( ch2 == APOSTROPHE ) || ( ch2 == RDQUOTE ) || ( ch2 == RSQUOTE ) ) {
				ch = MMGetChar( &MMFile );
				ch2 = MMPeekAhead( &MMFile, 0 );
			}

			int MaxCRs = 0, MaxLFs = 0;
			if ( FileIsPrewrapped )
				MaxCRs = MaxLFs = 1;

			while ( ( ch2 == SPACE ) || ( ch2 == NBSP ) || ( MaxCRs && ( ch2 == CR ) ) || ( MaxLFs && ( ch2 == LF ) ) ) {
				ch = MMGetChar( &MMFile );
				ch2 = MMPeekAhead( &MMFile, 0 );

				if ( ch2 == CR )
					MaxCRs--;
				else if ( ch2 == LF )
					MaxLFs--;
			}

			ch2 = MMPeekAhead( &MMFile, 0 );

			if ( ( ch2 != -1 ) && !iswlower( ch2 ) ) {

				for ( int i = MAX_CONTEXT_SENTENCES - 2; i >= 0; i-- )
					SentenceStartIndex[i + 1] = SentenceStartIndex[i];

				SentenceStartIndex[0] = MMFile.Index;

				if ( ContextCountdown ) {
					DumpEndIndex = MMFile.Index;
					ContextCountdown--;
				}

				ResetWordSeriesSearches( FirstWord );
				BlankLinesCount = 0;
			}
		}

		if ( _MMIsEOL( ch ) ) {
			if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) )
				MMGetChar( &MMFile );

			if ( BlankLine ) {
				InSingleQuotes = FALSE;
				BlankLinesCount++;
			}

			if ( BlankLinesCount == BlankLinesReset ) {

				for ( int i = MAX_CONTEXT_SENTENCES - 2; i >= 0; i-- )
					SentenceStartIndex[i + 1] = SentenceStartIndex[i];

				SentenceStartIndex[0] = MMFile.Index;

				if ( ContextCountdown ) {
					DumpEndIndex = MMFile.Index;
					ContextCountdown--;
				}

				ResetWordSeriesSearches( FirstWord );
				BlankLinesCount = 0;
			}

			BlankLine = TRUE;
		}
	}

	if ( ContextCountdown ) {
		if ( CTX_DumpBuffer( &MMFile, DumpStartIndex, MMFile.FileSize, Width, ConHandle, MoreArgs->NormalColor, MoreArgs->HighlightColor, &FirstFind ) == USER_ABORT )
			goto UserAbort;
	}
	else if ( DumpEndIndex ) {
		if ( CTX_DumpBuffer( &MMFile, DumpStartIndex, DumpEndIndex, Width, ConHandle, MoreArgs->NormalColor, MoreArgs->HighlightColor, &FirstFind ) == USER_ABORT )
			goto UserAbort;
	}

	if ( FoundWords ) {
		if ( ShowReport ) {
			if ( ContextShowFileReport( FirstWord, Filename ) == USER_ABORT )
				goto UserAbort;
		}
		else {
			if ( CL_EndLine() == USER_ABORT )
				goto UserAbort;
		}
	}
	else {
		QPuts( L"  No matches found." );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}

	FH_FreeLinkedList( (void**) &FirstFind );
	ContextFilesProcessed++;
	CloseMMFile( &MMFile );
	CloseHandle( ConHandle );
	CL_ReleaseBreak();
	return 0;


UserAbort:

	FH_FreeLinkedList( (void**) &FirstFind );
	CloseMMFile( &MMFile );
	CloseHandle( ConHandle );
	CL_ReleaseBreak();
	return USER_ABORT;
}



int CopyWord( LPTSTR inWord, LPTSTR outWord, BOOL AllowSpaces )
{
	if ( !inWord || !outWord )
		return -1;


	int j		= 0;
	outWord[j]	= '\0';
	wchar_t ch	= '\0';
	int Alnum	= 0;


	for ( int i = 0; inWord[i]; i++ ) {
		ch = towlower( StripAccent( inWord[i] ) );

		if ( AllowSpaces && ( ( ch == ';' ) || ( ch == ',' ) || ( ch == '+' ) ) )
			ch = SPACE;

		if ( iswspace( ch ) ) {
			ch = SPACE;
			if ( j == 0 )
				continue;
			else if ( outWord[j - 1] == SPACE )
				continue;
		}

		if ( iswalnum( ch ) )
			Alnum++;
		else if ( ( ch != APOSTROPHE ) && ( ch != '.' ) && ( ch != HYPHEN ) && ( ch != SPACE ) )
			return -1;

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			outWord[j++] = ch;
			outWord[j] = '\0';
		}
	}

	if ( !Alnum )
		return -1;

	if ( j )
		if ( outWord[j - 1] == SPACE )
			outWord[--j] = '\0';

	if ( j == 0 )
		return -1;

	if ( ( outWord[0] == '.' ) || ( outWord[0] == HYPHEN ) )
		return -1;

	if ( !AllowSpaces )
		if ( wcschr( outWord, SPACE ) != NULL )
			return -1;

	return 0;
}



DLLExports int WINAPI context( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t WordBuffer[SHORT_BUF_LEN]			= L"";

	int Flags					= 0;
	int Recurse					= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= NULL;
	BOOL NoHighlight			= FALSE;
	unsigned int UserHighlight	= 0;
	unsigned int Pause			= 0;

	SearchWord *FirstWord			= NULL;
	SearchWord *LastWord			= NULL;
	BOOL FilenameSpecified			= FALSE;
	BOOL StdinSpecified				= FALSE;
	BOOL WriteBOM					= TRUE;
	unsigned int SentencesBefore	= 1;
	unsigned int SentencesAfter		= 1;
	unsigned int Width				= 0;
	unsigned int UserFileFormat		= FMT_UNKNOWN;
	FindWordsArgs PassArgs ;
	int rv							= 0;
	int fhrv						= 0;


	memset( &PassArgs, 0x00, sizeof( FindWordsArgs ) );
	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	ContextFilesProcessed = 0;


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( (void**) &FirstWord );
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( ContextHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( CONTEXT_COMMAND_VERSION ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/CA", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &SentencesAfter, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad SentencesAfter \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( SentencesAfter >= MAX_CONTEXT_SENTENCES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"SentencesAfter must be 0 - %u\r\n", MAX_CONTEXT_SENTENCES - 1 );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/CB", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &SentencesBefore, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad SentencesBefore \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( SentencesBefore >= MAX_CONTEXT_SENTENCES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"SentencesBefore must be 0 - %u\r\n", MAX_CONTEXT_SENTENCES - 1 );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(PassArgs.CodePage), &(PassArgs.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &SentencesBefore, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad ContextSentences \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( SentencesBefore >= MAX_CONTEXT_SENTENCES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"ContextSentences must be 0 - %u\r\n", MAX_CONTEXT_SENTENCES - 1 );
					goto ErrorExit;
				}

				SentencesAfter = SentencesBefore;
			}
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &UserFileFormat, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad FileFormat \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( UserFileFormat > FMT_MAX_VALUE ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"FileFormat must be 0 - %u\r\n", FMT_MAX_VALUE );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/H", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' )
					UserHighlight = 0;
				else {
					unsigned int normal = 0, inverse = 0;
					int fg = 0, bg = 0, ufg = 0, ubg = 0;

					GetAttribute( &normal, &inverse );
					fg = normal & 0x0f;
					bg = ( normal & 0xf0 ) >> 4;

					if ( !_wcsnicmp( arg[i] + j, L"inv", 3 ) )
						ubg = ( fg & 0x07 ), ufg = bg;
					else if ( !_wcsnicmp( arg[i] + j, L"0x", 2 ) ) {
						unsigned int t = 0;
						if ( ParseInt( arg[i] + j, &t, NULL ) ) {
							DisplayErrorHeader();
							_wcsupr_s( arg[i], MAX_ARG_LEN );
							Qprintf( ERRHANDLE, L"Bad highlight color (bad hex):  \"%s\"\r\n", arg[i] + j );
							goto ErrorExit;
						}

						ufg = t & 0x0f;
						ubg = t >> 4;
					}
					else
						ParseColors( arg[i] + j, &ufg, &ubg );
					
					if ( ( ubg == 64 ) && ( ufg >= 0 ) && ( ufg <= 15 ) )
						ubg = -1;

					if ( ( ubg == ufg ) || ( ( ubg == -1 ) && ( ufg == bg ) ) ) {
						DisplayErrorHeader();
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						Qprintf( ERRHANDLE, L"Bad highlight color (invisible):  \"%s\"\r\n", arg[i] + j );
						goto ErrorExit;
					}

					if ( ( ufg >= 0 ) && ( ubg >= 0 ) )
						UserHighlight = ( ubg << 4 ) | ufg;
					else if ( ( ufg >= 0 ) && ( ubg == -1 ) )
						UserHighlight = ( bg << 4 ) | ufg;
					else {
						DisplayErrorHeader();
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						Qprintf( ERRHANDLE, L"Bad highlight color:  \"%s\"\r\n", arg[i] + j );
						goto ErrorExit;
					}
				}
			}
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				Width = ConsoleWidth();
			else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Width, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad width \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Width must be %u - %u\r\n", MIN_COLUMNS, MAX_COLUMNS );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				Flags |= CONTEXT_FLAG_REPORT;
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( CopyWord( arg[i] + j, WordBuffer, TRUE ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid word: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( AddSearchWord( &FirstWord, &LastWord, wcschr( WordBuffer, SPACE ) ? WORD_TYPE_SERIES : WORD_TYPE_START, WordBuffer ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/X", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( CopyWord( arg[i] + j, WordBuffer, TRUE ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid word: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( AddSearchWord( &FirstWord, &LastWord, wcschr( WordBuffer, SPACE ) ? WORD_TYPE_SERIES_EXACT : WORD_TYPE_EXACT, WordBuffer ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/Y", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( CopyWord( arg[i] + j, WordBuffer, FALSE ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid word: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				wchar_t temp[SHORT_BUF_LEN] = L"";
				Metaphone( WordBuffer, temp, CONTEXT_METAPHONE_LENGTH, 0 );
				wcscat_s( temp, SHORT_BUF_LEN, L":" );
				wcscat_s( temp, SHORT_BUF_LEN, WordBuffer );

				if ( AddSearchWord( &FirstWord, &LastWord, WORD_TYPE_METAPHONE, temp ) )
					goto ErrorExit;
			}
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}


	if ( !FirstWord ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No search words specified\r\n" );
		goto ErrorExit;
	}



	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( !Width ) {
		if ( ConsoleHandle( STD_OUTPUT_HANDLE ) )
			Width = ConsoleWidth();
		else
			Width = EnvironmentNumericValue( L"COLUMNS", MIN_COLUMNS, MAX_COLUMNS );

		if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) )
			Width = DEFAULT_COLUMNS;
	}


	PassArgs.FirstWord			= FirstWord;
	PassArgs.UserFormat			= UserFileFormat;
	PassArgs.SentencesBefore	= SentencesBefore;
	PassArgs.SentencesAfter		= SentencesAfter;
	PassArgs.Width				= Width;
	PassArgs.NormalColor		= CL_Normal;
	PassArgs.HighlightColor		= CL_Highlight;

	if ( CL_Normal && CL_Highlight && UserHighlight )
		PassArgs.HighlightColor	= UserHighlight;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = FindWords( (LPTSTR) StdinFilename, Flags, &PassArgs );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &PassArgs, FindWords, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	if ( ( Flags & CONTEXT_FLAG_REPORT ) && ( ContextFilesProcessed > 1 ) )
		if ( ContextShowFinalReport( FirstWord ) == USER_ABORT )
			goto UserAbort;


	FH_FreeLinkedList( (void**) &FirstWord );
	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( (void**) &FirstWord );
	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( (void**) &FirstWord );
	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	CONTEXT_COMMAND_VERSION


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				FindWords() filename is defined as a constant:  fix for mmfiles 1.4.4.2+

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
