
//			DEDUP command:


#ifndef DEDUP_COMMAND_VER
#define DEDUP_COMMAND_VER			1.0.1.1


#define DEDUP_MODE_MERGE			0x0000
#define DEDUP_MODE_UNIQUE			0x0001
#define DEDUP_MODE_REPEATED			0x0002
#define DEDUP_MODE_BITS				0x0003

#define DEDUP_IGNORE_CASE			0x0010
#define DEDUP_TRIM					0x0020
#define DEDUP_SKIP_BLANKS			0x0040
#define DEDUP_SHOW_HEADER			0x0080
#define DEDUP_SHOW_COUNTS			0x0100



const wchar_t DedupHelpText[] = L"Dump text files to stdout, merging repeated lines.\n\
\n\
DEDUP /B /C /CP:n /D /H /I /M /S /T /U filename....\r\n\
\n\
\t/B    discard blank lines\n\
\t/C    show line repeat counts\n\
\t/CP:  interpret OEM text as code page n\n\
\t/D    show only repeating lines\n\
\t/H    display filenames\n\
\t/I    ignore case when comparing lines\n\
\t/M    merge repeating lines (default)\n\
\t/S    search in subdirectories for matching files\n\
\t/T    trim leading and trailing whitespace\n\
\t/U    show only lines which do not repeat\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into DEDUP.\n";



struct _DedupArgs {
	unsigned int SkipFields;
	unsigned int SkipChars;
	unsigned int CompareMax;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



LPTSTR DD_SkipFields( LPTSTR inString, unsigned int Count )
{
	if ( !inString )
		return inString;
	
	if ( ( inString[0] == '\0' ) || !Count )
		return inString;


	wchar_t *ch = inString;

	for ( unsigned int i = 0; i < Count; i++ ) {
		while ( *ch && iswspace( *ch ) )
			ch++;

		while ( *ch && !iswspace( *ch ) )
			ch++;
	}

	return ch;
}


LPTSTR DD_SkipChars( LPTSTR inString, unsigned int Count )
{
	if ( !inString )
		return inString;

	if ( ( inString[0] == '\0' ) || !Count )
		return inString;


	wchar_t *ch = inString;

	for ( unsigned int i = 0; i < Count; i++ )
		if ( *ch )
			ch++;

	return ch;
}



int DedupFile( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t ThisLine[BIG_LINE_LEN]		= L"";
	wchar_t LastLine[BIG_LINE_LEN]		= L"";
	

	_MMFile MMFile;

	int Mode					= Flags & DEDUP_MODE_BITS;
	BOOL IgnoreCase				= ( Flags & DEDUP_IGNORE_CASE ) != 0;
	BOOL TrimWhitespace			= ( Flags & DEDUP_TRIM ) != 0;
	BOOL SkipBlanks				= ( Flags & DEDUP_SKIP_BLANKS ) != 0;
	BOOL ShowHeader				= ( Flags & DEDUP_SHOW_HEADER ) != 0;
	BOOL ShowCounts				= ( Flags & DEDUP_SHOW_COUNTS ) != 0;
	BOOL FirstLine				= TRUE;
	BOOL ShowLine				= FALSE;

	unsigned int SkipFields		= 0;
	unsigned int SkipChars		= 0;
	unsigned int CompareMax		= 0;
	unsigned int RepeatCount	= 0;
	LPTSTR String1				= NULL;
	LPTSTR String2				= NULL;

	_DedupArgs *args			= (_DedupArgs*) Args;
	if ( args ) {
		SkipFields	= args->SkipFields;
		SkipChars	= args->SkipChars;
		CompareMax	= args->CompareMax;
	}



	int rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return 1;


	CL_HookBreak();


	if ( ShowHeader )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;

	
	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, ThisLine, BIG_LINE_LEN, SkipBlanks ) )
			break;

		if ( TrimWhitespace )
			TrimLine( ThisLine );

		String1 = ThisLine;
		String2 = LastLine;

		if ( SkipFields ) {
			String1 = DD_SkipFields( String1, SkipFields );
			String2 = DD_SkipFields( String2, SkipFields );
		}

		if ( SkipChars ) {
			String1 = DD_SkipChars( String1, SkipChars );
			String2 = DD_SkipChars( String2, SkipChars );
		}


		if ( FirstLine )
			rv = 1;
		else if ( CompareMax ) {
			if ( IgnoreCase )
				rv = _wcsnicmp( String1, String2, CompareMax );
			else
				rv = wcsncmp( String1, String2, CompareMax );
		}
		else {
			if ( IgnoreCase )
				rv = _wcsicmp( String1, String2 );
			else
				rv = wcscmp( String1, String2 );
		}


		ShowLine = FALSE;

		switch ( Mode ) {

		case DEDUP_MODE_UNIQUE :		//  show only lines which are not repeated

			if ( rv && ( RepeatCount == 1 ) )
				ShowLine = TRUE;

			break;

		case DEDUP_MODE_REPEATED :		//  show only repeated lines

			if ( rv && ( RepeatCount > 1 ) )
				ShowLine = TRUE;

			break;

		default :						//  merge repeated lines

			if ( rv && !FirstLine )
				ShowLine = TRUE;
		}


		if ( ShowLine ) {
			if ( ShowCounts )
				Printf( L"%7u ", RepeatCount );

			QPuts( LastLine );
			if ( CL_EndLine() == USER_ABORT )
				goto UserAbort;
		}


		if ( rv ) {
			RepeatCount = 1;
			wcscpy_s( LastLine, MAX_ARG_LEN, ThisLine );
		}
		else
			RepeatCount++;

		FirstLine = FALSE;
	}


	CloseMMFile( &MMFile );


	ShowLine = FALSE;

	switch ( Mode ) {

	case DEDUP_MODE_UNIQUE :

		if ( !FirstLine && ( RepeatCount == 1 ) )
			ShowLine = TRUE;

		break;

	case DEDUP_MODE_REPEATED :

		if ( RepeatCount > 1 )
			ShowLine = TRUE;

		break;

	default :

		if ( RepeatCount && !FirstLine )
			ShowLine = TRUE;
	}


	if ( ShowLine ) {
		if ( ShowCounts )
			Printf( L"%7u ", RepeatCount );

		QPuts( LastLine );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}

	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI dedup( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	_DedupArgs Args			= { 0, 0, 0, 0, 0 };
	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( DedupHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( DEDUP_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Flags |= DEDUP_SKIP_BLANKS;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= DEDUP_SHOW_COUNTS;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Flags = ( Flags & ~DEDUP_MODE_BITS ) | DEDUP_MODE_REPEATED;
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Args.SkipFields, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad SkipFields value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= DEDUP_SHOW_HEADER;
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				Flags |= DEDUP_IGNORE_CASE;
			else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Args.SkipChars, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad SkipChars number: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/M" ) )
				Flags = ( Flags & ~DEDUP_MODE_BITS ) | DEDUP_MODE_MERGE;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Flags |= DEDUP_TRIM;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				Flags = ( Flags & ~DEDUP_MODE_BITS ) | DEDUP_MODE_UNIQUE;
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Args.CompareMax, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad CompareMax value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = DedupFile( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, DedupFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	DEDUP_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				DedupFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
