
//			FILTERFILES command:


#ifndef FILTERFILES_COMMAND_VER
#define FILTERFILES_COMMAND_VER				1.0.1.0


#define FILTERFILES_FLAG_NO_COLOR			0x0001
#define FILTERFILES_FLAG_QUIETLY			0x0002
#define FILTERFILES_FLAG_IGNORE_EXIT_3		0x0004
#define FILTERFILES_FLAG_REDIRECT_INPUT		0x0008
#define FILTERFILES_FLAG_NOT_REALLY			0x0010



const wchar_t FFC_BackupExt[]		= L".original";


const wchar_t FilterFilesHelpText[] = L"Pass files through a text filter command.\n\
\n\
FILTERFILES /B:.ext /C /N /J /P /Q /S /UTF8 /UTF16 filespec.... : command args\r\n\
\n\
\t/B:     backup extension (default is .original)\n\
\t/C      do not abort if command exits with errorlevel 3\n\
\t/N      not really\n\
\t/J      redirect input\n\
\t/P      prompt first\n\
\t/Q      quietly\n\
\t/S      search in subdirectories for matching files\n\
\t/UTF8   redirect output as UTF-8\n\
\t/UTF16  redirect output as UTF-16\n\
\n\
At least one filespec is required.  Anything after the first unquoted\n\
colon is the command to execute; this also is required.\n\
\n\
Matching files will be renamed with a .original extension (or as per /B).\n\
Then the specified command will be called, passing the new filename, and\n\
with output redirected to the original filename.\n\
\n\
This command only processes local files.  CLIP:, URLs, standard input,\n\
etc. are not supported.\n";



struct _FilterFileArgs {
	LPTSTR Command;
	LPTSTR BackupExt;
	BOOL Prompt;
	unsigned int Found;
	unsigned int Processed;
	unsigned int Skipped;
	unsigned int Failed;
};




BOOL FFC_OkayExtension( LPTSTR String )
{
	if ( !String )
		return FALSE;


	if ( ( String[0] != '.' ) || ( String[1] == '\0' ) )
		return FALSE;


	for ( wchar_t *ch = String + 1; *ch; ch++ ) {
		if ( ( *ch == '.' ) || ( *ch == '/' ) || ( *ch == '\\' ) )
			return FALSE;

		if ( ( *ch == '?' ) || ( *ch == '*' ) )
			return FALSE;

		if ( iswcntrl( *ch ) )
			return FALSE;
	}

	return TRUE;
}



LPTSTR FFC_FindMarker( LPTSTR String, LPTSTR Marker )
{
	if ( !String || !Marker )
		return NULL;
	else if ( ( String[0] == '\0' ) || ( Marker[0] == '\0' ) )
		return NULL;


	wchar_t *ch			= String;
	BOOL quote			= FALSE;
	BOOL WordStart		= TRUE;


	while ( *ch ) {

		if ( *ch == QUOTE )
			quote = !quote;
		else {
			if ( WordStart && !quote && ( *ch == Marker[0] ) )
				if ( !wcsncmp( ch, Marker, wcslen( Marker ) ) )
					return ch;

			WordStart = iswspace( *ch );
		}

		ch++;
	}

	return NULL;
}



int FFC_GetFirstWord( LPTSTR inString, LPTSTR outWord, size_t buflen )
{
	if ( !inString || !outWord || !buflen )
		return -1;


	wchar_t *ch			= inString;
	size_t j			= 0;
	BOOL quote			= FALSE;
	int InPats			= 0;


	outWord[j] = '\0';

	while ( *ch ) {

		if ( *ch == QUOTE )
			quote = !quote;
		else if ( ( *ch == '(' ) && !quote ) {
			InPats++;
			j = 0;
			outWord[j] = '\0';
			ch++;
			continue;
		}
		else if ( iswspace( *ch ) && !quote ) {
			if ( j )
				return 0;

			ch++;
			continue;
		}
		else if ( ( *ch == ')' ) && !quote && !InPats )
			return 0;

		if ( j < ( buflen - 1 ) ) {
			outWord[j++] = *ch;
			outWord[j] = '\0';
		}

		ch++;
	}

	return 0;
}



int FFC_SearchAppPaths( LPTSTR inFilename, HKEY RootKey, LPTSTR outFullname )
{
	if ( !inFilename || !outFullname || !RootKey )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	wchar_t KeyName[MAX_FILENAME_LEN]	= L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\";
	wchar_t Data[MAX_FILENAME_LEN]		= L"";
	HKEY KeyHandle		= NULL;
	DWORD DataType = 0, DataSize = MAX_FILENAME_LEN * sizeof( wchar_t );
	int rv = 0;


	wcscat_s( KeyName, MAX_FILENAME_LEN, inFilename );

	if ( RegOpenKeyEx( RootKey, KeyName, 0, KEY_READ, &KeyHandle ) )
		return 1;

	rv = RegQueryValueEx( KeyHandle, NULL, NULL, &DataType, (BYTE *) Data, &DataSize );

	RegCloseKey( KeyHandle );

	if ( rv != ERROR_SUCCESS )
		return 1;

	if ( ( DataType != REG_SZ ) && ( DataType != REG_EXPAND_SZ ) )
		return 1;

	if ( ( DataType == REG_EXPAND_SZ ) && ( wcschr( Data, '%' ) != NULL ) ) {
		wchar_t temp[MAX_FILENAME_LEN] = L"";
		ExpandEnvironmentStrings( Data, temp, MAX_FILENAME_LEN );
		wcscpy_s( Data, MAX_FILENAME_LEN, temp );
	}

	if ( QueryIsFile( Data ) == 0 )
		return 1;

	wcscpy_s( outFullname, MAX_FILENAME_LEN, Data );
	return 0;
}



int FFC_SearchPathForExe( LPTSTR inFilename, LPTSTR outFullname, BOOL TryBatches, LPTSTR Extension = NULL )
{
	if ( !inFilename || !outFullname )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	LPTSTR Ext							= NULL;
	BOOL DoRegSearch					= FALSE;


	outFullname[0] = '\0';


	wcscpy_s( temp, MAX_FILENAME_LEN, inFilename );
	TrimQuotes( temp );


	Ext = FH_ExtensionPart( temp, NULL );

	if ( !Ext && !Extension ) {

#ifndef _WIN64
		if ( FFC_SearchPathForExe( temp, outFullname, FALSE, L".com" ) == 0 )
			return 0;
#endif

		if ( FFC_SearchPathForExe( temp, outFullname, FALSE, L".exe" ) == 0 )
			return 0;

		if ( TryBatches ) {
			if ( FFC_SearchPathForExe( temp, outFullname, FALSE, L".btm" ) == 0 )
				return 0;

			if ( FFC_SearchPathForExe( temp, outFullname, FALSE, L".cmd" ) == 0 )
				return 0;

			if ( FFC_SearchPathForExe( temp, outFullname, FALSE, L".bat" ) == 0 )
				return 0;
		}
	}


	if ( Extension && !Ext ) {
		wcscat_s( temp, MAX_FILENAME_LEN, Extension );

		if ( !_wcsicmp( Extension, L".exe" ) )
			DoRegSearch = TRUE;
	}
	else if ( Ext ) {
		if ( !_wcsicmp( Ext, L".exe" ) )
			DoRegSearch = TRUE;
	}


	if ( DoRegSearch ) {
		if ( FFC_SearchAppPaths( temp, HKEY_CURRENT_USER, outFullname ) == 0 )
			return 0;

		if ( FFC_SearchAppPaths( temp, HKEY_LOCAL_MACHINE, outFullname ) == 0 )
			return 0;
	}

	SearchPath( NULL, temp, NULL, MAX_FILENAME_LEN, outFullname, NULL );

	if ( outFullname[0] )
		return 0;

	return 1;
}



int FFC_CommandType( LPTSTR Command )
{
	if ( !Command )
		return -1;
	else if ( Command[0] == '\0' )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	LPTSTR Command1						= Command;


	if ( ( Command[0] == '*' ) && ( Command[1] ) )
		Command1++;


	swprintf( temp, MAX_FILENAME_LEN, L"%%@if[isalias %s,1,0]", Command );
	ExpandVariables( temp, 1 );
	if ( temp[0] == '1' )
		return 1;

	swprintf( temp, MAX_FILENAME_LEN, L"%%@if[isplugin %s,2,0]", Command1 );
	ExpandVariables( temp, 1 );
	if ( ( temp[0] == '2' ) && ( Command[0] != '%' ) )
		return 2;

	swprintf( temp, MAX_FILENAME_LEN, L"%%@if[isinternal %s,3,0]", Command1 );
	ExpandVariables( temp, 1 );
	if ( temp[0] == '3' )
		return 3;

	temp[0] = '\0';
	FFC_SearchPathForExe( Command1, temp, TRUE );
	if ( temp[0] )
		if ( QueryIsFile( temp ) )
			return 4;

	return 0;
}



int FilterFile( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename || !Args )
		return -1;


	wchar_t CmdBuf[BIG_LINE_LEN]		= L"";
	wchar_t OldName[MAX_FILENAME_LEN]	= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	_FilterFileArgs *args	= (_FilterFileArgs*) Args;

	LPTSTR Ext				= NULL;
	BOOL NoColor			= Flags & FILTERFILES_FLAG_NO_COLOR;
	BOOL Quietly			= Flags & FILTERFILES_FLAG_QUIETLY;
	BOOL IgnoreExitCode3	= Flags & FILTERFILES_FLAG_IGNORE_EXIT_3;
	BOOL RedirectInput		= Flags & FILTERFILES_FLAG_REDIRECT_INPUT;
	BOOL NotReally			= Flags & FILTERFILES_FLAG_NOT_REALLY;
	int rv					= 0;



	wcscpy_s( OldName, MAX_FILENAME_LEN, Filename );
	Ext = FH_ExtensionPart( OldName, NULL );

	if ( Ext ) {
		if ( !_wcsicmp( Ext, args->BackupExt ) )
			return 0;

		Ext[0] = '\0';
	}

	wcscat_s( OldName, MAX_FILENAME_LEN, args->BackupExt );
	args->Found += 1;


	if ( QueryIsFile( OldName ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Skipping \"%s\"\r\n", Filename );
		Qprintf( ERRHANDLE, L"   because \"%s\" exists.\r\n\r\n", OldName );
		args->Skipped += 1;
		return ERROR_FILE_EXISTS;
	}


	if ( !Quietly || args->Prompt ) {
		CL_NoPausing();
		CL_PickHighlightColor( NoColor );
		HLprintf( L"\x12%s \x11%s\"%s\" > \"%s\"\n", args->Command, RedirectInput ? L"< " : L"", OldName, Filename );
	}


	if ( args->Prompt ) {
		int ans = PromptUser2( L"   Filter this file?  [Y/N/A/Q] " );

		if ( ans == USER_ABORT ) {
			HLputs( L"\r\n   * User abort *\r\n\r\n" );
			return USER_ABORT;
		}
		else if ( ans == 0 ) {
			QPuts( L"Skipped.\r\n\r\n" );
			args->Skipped += 1;
			return 0;
		}
		else if ( ans == 2 )
			args->Prompt = FALSE;
	}


	if ( NotReally ) {
		if ( !Quietly )
			QPuts( L"   Exit code: 0\r\n\r\n" );

		args->Processed += 1;
		return 0;
	}


	if ( !MoveFile( Filename, OldName ) ) {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error renaming file \"%s\":\r\n", Filename );
		Qprintf( ERRHANDLE, L"   %s\r\n\r\n", ErrorText );
		args->Skipped += 1;
		return 1;
	}

	swprintf( CmdBuf, BIG_LINE_LEN, L"%s %s\"%s\" > \"%s\"", args->Command, RedirectInput ? L"< " : L"", OldName, Filename );
	//	wprintf( L"   * CmdBuf = %s\n", CmdBuf );


	args->Processed += 1;

	try {
		rv = Command( CmdBuf, 0 );
	}
	catch ( ... ) {
		rv = USER_ABORT;
	}

	if ( !Quietly )
		Printf( L"   Exit code: %d\r\n\r\n", rv );


	if ( !IgnoreExitCode3 && ( rv == 3 )  )
		rv = USER_ABORT;

	if ( rv )
		args->Failed += 1;

	return rv;
}



DLLExports int WINAPI filterfiles( LPTSTR inString )
{
	if ( !inString )
		return 1;


	const int MaxFilespecs					= 128;
	wchar_t Filespec[MaxFilespecs][MAX_ARG_LEN];

	wchar_t Marker[]						= L":";
	LPTSTR FilterCmdStart					= NULL;
	wchar_t FilterCmd[MAX_FILENAME_LEN]		= L"";
	wchar_t CommandWord[MAX_FILENAME_LEN]	= L"";

	wchar_t BackupExt[SHORT_BUF_LEN]		= L"";


	int Flags					= 0;
	int Recurse					= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= NULL;
	BOOL NoHighlight			= FALSE;

	int CommandType				= 0;
	LPTSTR Ext					= NULL;
	int OldUnicodeOut			= 0;
	int UserUnicodeOut			= 0;

	int NumFilespecs			= 0;
	BOOL Verbose				= FALSE;
	BOOL Advanced				= FALSE;
	BOOL Prompt					= FALSE;
	BOOL BadFilespec			= FALSE;
	int rv						= 0;
	int fhrv					= 0;

	HANDLE OldStdoutHandle		= NULL;
	HANDLE OldStderrHandle		= NULL;
	HANDLE NewConsoleHandle		= NULL;
	BOOL FixedStdout			= FALSE;
	BOOL FixedStderr			= FALSE;

	_FilterFileArgs Args;



	OldUnicodeOut = OutputIsUnicode();

	FilterCmdStart = FFC_FindMarker( inString, Marker );

	if ( FilterCmdStart ) {
		FilterCmdStart[0] = '\0';
		FilterCmdStart += wcslen( Marker );
		while ( iswspace( FilterCmdStart[0] ) )
			FilterCmdStart++;
	}



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( FilterFilesHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( FILTERFILES_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/B", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				BackupExt[0] = '\0';

				if ( FFC_OkayExtension( arg[i] + j ) )
					wcscpy_s( BackupExt, SHORT_BUF_LEN, arg[i] + j );

				if ( BackupExt[0] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], SHORT_BUF_LEN );
					Qprintf( ERRHANDLE, L"Bad extension: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= FILTERFILES_FLAG_IGNORE_EXIT_3;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = 1;
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				Flags |= FILTERFILES_FLAG_NOT_REALLY;
			else if ( !_wcsicmp( arg[i], L"/J" ) )
				Flags |= FILTERFILES_FLAG_REDIRECT_INPUT;
			else if ( !_wcsicmp( arg[i], L"/OEM" ) )
				UserUnicodeOut = 1;
			else if ( !_wcsicmp( arg[i], L"/P" ) )
				Prompt = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= FILTERFILES_FLAG_QUIETLY;
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;

				if ( !_wcsnicmp( arg[i] +j, L"TF", 2 ) )
					j += 2;

				if ( arg[i][j] == '-' )
					j++;

				if ( !_wcsicmp( arg[i] + j, L"8" ) )
					UserUnicodeOut = 8;
				else if ( !_wcsicmp( arg[i] + j, L"16" ) )
					UserUnicodeOut = 16;
				else {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Unknown suboption: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				Verbose = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				Advanced = TRUE;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, &NoHighlight, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( NumFilespecs >= MaxFilespecs ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Too many filespecs!  (%d max)\r\n", MaxFilespecs );
					goto ErrorExit;
				}

				wcscpy_s( Filespec[NumFilespecs], MAX_ARG_LEN, arg[i] );
				NumFilespecs++;
			}
		}
	}


	if ( FilterCmdStart ) {
		wcscpy_s( FilterCmd, MAX_FILENAME_LEN, FilterCmdStart );
		FFC_GetFirstWord( FilterCmd, CommandWord, MAX_FILENAME_LEN );
	}

	if ( ( CommandWord[0] == '\0' ) && !Advanced ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error: Missing %s command\r\n", Marker );
		goto ErrorExit;
	}

	if ( !Advanced ) {
		if ( wcschr( FilterCmd, '<' ) || wcschr( FilterCmd, '>' ) || wcschr( FilterCmd, '|' ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Command should not contain redirections: \"%s\"\r\n", FilterCmd );
			goto ErrorExit;
		}
	}


	if ( !NumFilespecs ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No filename specified\r\n" );
		goto ErrorExit;
	}


	if ( ShellTccLe && ( UserUnicodeOut == 8 ) )		//	//UTF8Output=Yes does not actually work in TCC/LE
		UserUnicodeOut = 0;



	//	wprintf( L"   * FilterCmd = '%s'\r\n", FilterCmd );
	//	wprintf( L"   * CommandWord = '%s'\r\n", CommandWord );

	CommandType = FFC_CommandType( CommandWord );
	_wcsupr_s( CommandWord, MAX_FILENAME_LEN );

	if ( !CommandType && !Advanced ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Unknown command: \"%s\"\r\n", CommandWord );
		goto ErrorExit;
	}


	if ( Verbose ) {

		Printf( L"%s is ", CommandWord );

		switch ( CommandType ) {

		case 0 :
			QPuts( L"an unknown command.\r\n" );
			break;

		case 1 :
			QPuts( L"an alias.\r\n" );
			break;

		case 2 :
			QPuts( L"a plugin command.\r\n" );
			break;

		case 3 :
			QPuts( L"an internal command.\r\n" );
			break;

		case 4 :
			QPuts( L"an external command.\r\n" );
			break;

		default :
			QPuts( L"a mystery to me.\r\n" );
		}
	}


	if ( BackupExt[0] == '\0' )
		wcscpy_s( BackupExt, SHORT_BUF_LEN, FFC_BackupExt );



	for ( int i = 0; i < NumFilespecs; i++ ) {
		if ( !_wcsicmp( Filespec[i], L"clip:" ) ) {
			DisplayErrorHeader();
			_wcsupr_s( Filespec[i], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Not a valid filename: \"%s\"\r\n", Filespec[i] );
			BadFilespec = TRUE;
		}

		if ( QueryIsURL( Filespec[i] ) ) {
			DisplayErrorHeader();
			_wcsupr_s( Filespec[i], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"URLs not supported: \"%s\"\r\n", Filespec[i] );
			BadFilespec = TRUE;
		}

		Ext = FH_ExtensionPart( Filespec[i], NULL );
		if ( Ext ) {
			if ( !_wcsicmp( Ext, BackupExt ) ) {
				DisplayErrorHeader();
				_wcsupr_s( Filespec[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Filespec with backup extension: \"%s\"\r\n", Filespec[i] );
				BadFilespec = TRUE;;
			}
		}
	}

	if ( BadFilespec )
		goto ErrorExit;



	if ( UserUnicodeOut ) {
		fhrv = SetOutputEncoding( UserUnicodeOut );

		if ( fhrv == ERROR_NOT_SUPPORTED ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"This version of TCC does not support UTF-8 redirection\r\n" );
			goto ErrorExit;
		}
	}


	OldStdoutHandle		= GetStdHandle( STD_OUTPUT_HANDLE );
	OldStderrHandle		= GetStdHandle( STD_ERROR_HANDLE );

	if ( !ConsoleHandle( OldStdoutHandle ) || !ConsoleHandle( OldStderrHandle ) ) {
		NewConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

		if ( NewConsoleHandle != INVALID_HANDLE_VALUE ) {
			if ( !ConsoleHandle( OldStdoutHandle ) ) {
				SetStdHandle( STD_OUTPUT_HANDLE, NewConsoleHandle );
				FixedStdout = TRUE;
			}

			if ( !ConsoleHandle( OldStderrHandle ) ) {
				SetStdHandle( STD_ERROR_HANDLE, NewConsoleHandle );
				FixedStderr = TRUE;
			}
		}
		else
			NewConsoleHandle = NULL;
	}


	if ( NoHighlight )
		Flags |= FILTERFILES_FLAG_NO_COLOR;


	memset( &Args, 0x00, sizeof( _FilterFileArgs ) );

	Args.Command	= FilterCmd;
	Args.BackupExt	= BackupExt;
	Args.Prompt		= Prompt;


	for ( int i = 0; i < NumFilespecs; i++ ) {
		fhrv = HandleMatchingFiles( Filespec[i], Flags, &Args, FilterFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
		if ( fhrv == USER_ABORT )
			goto UserAbort;
		else if ( fhrv )
			rv = 2;
	}


	if ( ( Flags & FILTERFILES_FLAG_QUIETLY ) == 0 )
		Printf( L"%u file%s found, %u processed, %u skipped.  %u failed.\r\n", Args.Found, ( Args.Found == 1 ) ? L"" : L"s", Args.Processed, Args.Skipped, Args.Failed );


	goto CleanUpAndExit;


ErrorExit :

	rv = 1;
	goto CleanUpAndExit;


UserAbort :

	rv = 3;
	goto CleanUpAndExit;


CleanUpAndExit :

	FH_FreeLinkedList( &FirstRange );

	if ( UserUnicodeOut )
		SetOutputEncoding( OldUnicodeOut );

	if ( FixedStdout )
		SetStdHandle( STD_OUTPUT_HANDLE, OldStdoutHandle );

	if ( FixedStderr )
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );

	if ( NewConsoleHandle )
		CloseHandle( NewConsoleHandle );

	return rv;
}


#endif	//	FILTERFILES_COMMAND_VER


/*

	-			2021-10-05

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				FilterFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2

*/
