
//			UTYPE command:


#ifndef UTYPE_COMMAND_VERSION
#define UTYPE_COMMAND_VERSION		1.0.1.2



#define UTYPE_LINE_NUMBERS			0x00001
#define UTYPE_NO_BLANKS				0x00002
#define UTYPE_NO_BELLS				0x00004
#define UTYPE_HEADER				0x00008
#define UTYPE_REPLACE_CONTROLS		0x00010
#define UTYPE_TRIM					0x00020
#define UTYPE_NO_LEADING_BLANKS		0x00040
#define UTYPE_FIX_QUOTES			0x00080
#define UTYPE_CR_ONLY				0x00100
#define UTYPE_LF_ONLY				0x00200
#define UTYPE_HANDBRAKE				0x00400
#define UTYPE_FORCE_EBCDIC			0x00800
#define UTYPE_INCLUDE_AFTER_STRING	0x01000
#define UTYPE_INCLUDE_UNTIL_STRING	0x02000
#define UTYPE_HEX_SPACES			0x04000
#define UTYPE_HEX					0x08000
#define UTYPE_VERBOSE_HEADER		0x10000
#define UTYPE_SKIP_NULLS			0x40000
#define UTYPE_NULLS_TO_INVCHAR		0x80000




const wchar_t UtypeHelpText[] = L"Dump text files to standard output.\n\
\n\
UTYPE /B /C /CP:n /D /E /F:str /H /K:n /L /P /Q /S /T /U:str /X filename.... \r\n\
\n\
\t/B    no BEL characters\n\
\t/C    replace control characters\n\
\t/CP:  interpret OEM text as code page n\n\
\t/D    discard blank lines at start of file\n\
\t/E    discard all empty lines\n\
\t/F:   show only lines following this string; /FF: inclusive\n\
\t/H    display filenames\n\
\t/K:   expand tabs to n columns\n\
\t/L    line numbers\n\
\t/P    page output\n\
\t/Q    replace ASCII quotes and apostrophes\n\
\t/S    search in subdirectories for matching files\n\
\t/T    trim leading and trailing whitespace\n\
\t/U:   show only lines until (before) this string; /UU: inclusive\n\
\t/X    hex dump\n\
\n\
Input filenames may be specified on the command line, or text may\n\
be piped or redirected into UTYPE.\n";



struct TypeFileArgs {
	int MaxLength;
	unsigned int TabWidth;
	LPTSTR LineNumbersFormat;
	LPTSTR ShowAfterString;
	LPTSTR ShowUntilString;
	int HexWidth;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};





int StripBells( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	wchar_t ch = '\0';
	size_t i = 0, j = 0;

	do {
		ch = String[i++];
		if ( ch == 0x07 )
			continue;

		String[j++] = ch;

	} while ( ch );

	return 0;
}


int ReplaceControlChars( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	wchar_t temp[BIG_LINE_LEN] = L"";
	int j = 0;
	wchar_t ch = '\0';
	int rv = 0;

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch < 0x20 ) {
			if ( j < ( BIG_LINE_LEN - 1 ) ) {
				temp[j++] = '^';
				ch += 0x40;
			}
			else
				ch = '^';

			rv++;
		}
		else if ( ch == 0x7f ) {
			if ( j < ( BIG_LINE_LEN - 1 ) ) {
				temp[j++] = '^';
				ch = '?';
			}
			else
				ch = '^';

			rv++;
		}
		else if ( ( ch >= 0x80 ) && ( ch <= 0x9f ) ) {
			if ( j < ( BIG_LINE_LEN - 6 ) ) {
				temp[j++] = '\\';
				temp[j++] = 'u';
				temp[j++] = '0';
				temp[j++] = '0';
				temp[j++] = ( ch < 0x90 ) ? '8' : '9';
				ch = ch & 0x0f;
				if ( ch < 10 )
					ch = '0' + ch;
				else
					ch = 'A' + ch - 10;
			}
			else
				ch = '?';

			rv++;
		}

		if ( j < ( BIG_LINE_LEN - 1 ) ) {
			temp[j++] = ch;
			temp[j] = '\0';
		}
	}

	if ( !rv )
		return rv;

	wcscpy_s( String, BIG_LINE_LEN, temp );
	return rv;
}



int TypeFileHex( _MMFile *MMFile, int HexWidth, BOOL SpacesForControls, BOOL Handbrake )
{
	wchar_t OutBuffer[SHORT_BUF_LEN]	= L"";

	int ch					= 0;
	wchar_t ControlChar		= 0x00b7;
	int j					= 0;


	if ( SpacesForControls )
		ControlChar = SPACE;


	for ( size_t b = 0; b < MMFile->FileSize; b += HexWidth ) {
		//	wsprintf( OutBuffer, L"%04x %04x  ", b >> 16, b & 0xffff );
		wsprintf( OutBuffer, L"%08I64x  ", (unsigned __int64) b );

		j = (int) wcslen( OutBuffer );

		for ( int i = 0; i < HexWidth; i++ ) {
			if ( ( b + i ) < MMFile->FileSize ) {
				ch = MMFile->Buffer[b + i];
				wsprintf( OutBuffer + j, L"%02x ", ch );
				j += 3;
			}
			else {
				wcscat_s( OutBuffer, SHORT_BUF_LEN, L"   " );
				j += 3;
			}

			if ( ( i % 8 ) == 7 ) {
				wcscat_s( OutBuffer, SHORT_BUF_LEN, L" " );
				j++;
			}
		}


		for ( int i = 0; i < HexWidth; i++ ) {
			if ( ( b + i ) == MMFile->Index )
				ch = MMGetChar( MMFile );
			else
				ch = SPACE;

			if ( !ch || iswcntrl( ch ) )
				ch = ControlChar;

			if ( ch > 0 )
				OutBuffer[j++] = ch;
		}

		OutBuffer[j] = '\0';
		QPuts( OutBuffer );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;

		if ( Handbrake )
			CL_Handbrake();

	}

	CloseMMFile( MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



int TypeFile( LPCTSTR Filename, int Flags, void *TypeArgs )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t LineBuffer[BIG_LINE_LEN]	= L"";
	wchar_t OldLineVar[MAX_ARG_LEN]		= L"";
	wchar_t temp[BIG_LINE_LEN]			= L"";
	wchar_t VarName[]					= L"_LINE";
	unsigned int LineNumber				= 0;
	wchar_t ch							= SPACE;
	wchar_t lastch						= '\0';
	wchar_t nextch						= '\0';
	int j								= 0;
	BOOL ShowAfter						= FALSE;
	BOOL ShowUntil						= FALSE;
	BOOL SkipBlanks						= ( Flags & UTYPE_NO_BLANKS ) > 0;
	BOOL BugOut							= FALSE;
	int MMFlags							= MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC | MM_FLAG_HIGH_CHARS;

	_MMFile MMFile;

	TypeFileArgs *MoreArgs				= (TypeFileArgs*) TypeArgs;
	int MaxLength						= MoreArgs->MaxLength;
	unsigned int TabWidth				= MoreArgs->TabWidth;
	LPTSTR LineNumbersFormat			= MoreArgs->LineNumbersFormat;
	LPTSTR ShowAfterString				= MoreArgs->ShowAfterString;
	LPTSTR ShowUntilString				= MoreArgs->ShowUntilString;
	


	if ( ShowAfterString )
		if ( ShowAfterString[0] )
			ShowAfter = TRUE;

	if ( ShowUntilString )
		if ( ShowUntilString[0] )
			ShowUntil = TRUE;

	if ( Flags & UTYPE_HEX )
		Flags &= ~UTYPE_SKIP_NULLS;

	if ( Flags & UTYPE_SKIP_NULLS )
		MMFlags |= MM_FLAG_SKIP_NULLS;

	if ( Flags & UTYPE_NULLS_TO_INVCHAR )
		MMFlags |= MM_FLAG_NULLS_TO_INVCHAR;


	GetEnvironmentVariable( VarName, OldLineVar, MAX_ARG_LEN );
	

	//  Open file:

	int rv = OpenMMFile( &MMFile, Filename, MMFlags, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	if ( ( Flags & UTYPE_FORCE_EBCDIC ) && ( MMFile.FileEncoding == FILE_OEM ) )
		MMFile.FileEncoding = FILE_EBCDIC;


	CL_HookBreak();
	

	if ( Flags & UTYPE_HEADER ) {
		if ( ShowFilenameHeader( &MMFile, ( Flags & UTYPE_VERBOSE_HEADER ) > 0 ) == USER_ABORT )
			goto UserAbort;
	}

	if ( Flags & UTYPE_HEX )
		return TypeFileHex( &MMFile, MoreArgs->HexWidth, ( Flags & UTYPE_HEX_SPACES ) > 0, ( Flags & UTYPE_HANDBRAKE ) > 0 );


	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, SkipBlanks | MMGL_FLAG_SPLIT_LONG_LINES ) )
			break;

		if ( ShowAfter ) {
			if ( LCStringContains( LineBuffer, ShowAfterString ) ) {
				ShowAfter = FALSE;
			
				if ( ( Flags & UTYPE_INCLUDE_AFTER_STRING ) == 0 )
					continue;
			}
			else
				continue;
		}

		if ( ShowUntil ) {
			if ( LCStringContains( LineBuffer, ShowUntilString ) ) {
				if ( Flags & UTYPE_INCLUDE_UNTIL_STRING )
					BugOut = TRUE;
				else
					break;
			}
		}

		if ( Flags & UTYPE_NO_LEADING_BLANKS )
			SkipBlanks = FALSE;

		if ( Flags & UTYPE_TRIM )
			TrimLine( LineBuffer );

		if ( Flags & UTYPE_NO_BELLS )
			StripBells( LineBuffer );

		if ( Flags & UTYPE_REPLACE_CONTROLS ) {
			if ( TabWidth )
				ReplaceChar( LineBuffer, '\t', NON_CHARACTER );

			ReplaceControlChars( LineBuffer );

			if ( TabWidth )
				ReplaceChar( LineBuffer, NON_CHARACTER, '\t' );
		}

		if ( TabWidth )
			Detab( LineBuffer, TabWidth, SPACE );

		if ( Flags & UTYPE_FIX_QUOTES )
			FixQuotes( LineBuffer, BIG_LINE_LEN );

		if ( MaxLength )
			if ( (int) wcslen( LineBuffer ) > MaxLength )
				LineBuffer[MaxLength] = '\0';

		if ( Flags & UTYPE_LINE_NUMBERS ) {
			if ( !LineNumbersFormat )
				wsprintf( temp, L"%4u : ", LineNumber + 1 );
			else {
				wsprintf( temp, L"%u", LineNumber );
				SetEnvironmentVariable( VarName, temp );
				wcscpy_s( temp, BIG_LINE_LEN, (LPTSTR) LineNumbersFormat );
				ExpandVariables( temp, 1 );
			}

			wcscat_s( temp, BIG_LINE_LEN, LineBuffer );
			wcscpy_s( LineBuffer, BIG_LINE_LEN, temp );
		}

		QPuts( LineBuffer );

		if ( Flags & UTYPE_CR_ONLY )
			QPuts( L"\r" );
		else if ( Flags & UTYPE_LF_ONLY ) 
			QPuts( L"\n" );
		else if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;

		LineNumber++;

		if ( Flags & UTYPE_HANDBRAKE )
			CL_Handbrake();

		if ( BugOut )
			break;
	}
	
	CloseMMFile( &MMFile );

	if ( Flags & UTYPE_LINE_NUMBERS ) {
		if ( OldLineVar[0] == '\0' )
			SetEnvironmentVariable( VarName, NULL );
		else
			SetEnvironmentVariable( VarName, OldLineVar );
	}

	CL_ReleaseBreak();
	return 0;


UserAbort:

	CloseMMFile( &MMFile );

	if ( Flags & UTYPE_LINE_NUMBERS ) {
		if ( OldLineVar[0] == '\0' )
			SetEnvironmentVariable( VarName, NULL );
		else
			SetEnvironmentVariable( VarName, OldLineVar );
	}

	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI utype( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t ShowAfterString[MAX_ARG_LEN]	= L"";
	wchar_t ShowUntilString[MAX_ARG_LEN]	= L"";

	int Flags				= UTYPE_HANDBRAKE;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	LPTSTR LineNumberFormat	= NULL;
	unsigned int MaxLength	= 0;
	unsigned int TabWidth	= 0;
	int HexWidth			= 0;
	BOOL TabWidthSpecified	= FALSE;
	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	TypeFileArgs PassArgs	= { 0, 0, NULL, NULL, NULL, 0, 0, 0 };


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( UtypeHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( UTYPE_COMMAND_VERSION ) );
			}
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Flags |= UTYPE_NO_BELLS;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= UTYPE_REPLACE_CONTROLS;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(PassArgs.CodePage), &(PassArgs.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/CR" ) )
				Flags = Flags & ~UTYPE_LF_ONLY | UTYPE_CR_ONLY ;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Flags |= UTYPE_NO_BLANKS | UTYPE_NO_LEADING_BLANKS;
			else if ( !_wcsicmp( arg[i], L"/E" ) )
				Flags |= UTYPE_NO_BLANKS;
			else if ( !_wcsicmp( arg[i], L"/EBCDIC" ) )
				Flags |= UTYPE_FORCE_EBCDIC;
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;

				if ( towupper( arg[i][j] ) == 'F' ) {
					j++;
					Flags |= UTYPE_INCLUDE_AFTER_STRING;
				}

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( ShowAfterString, MAX_ARG_LEN, arg[i] + j );
				_wcslwr_s( ShowAfterString, MAX_ARG_LEN );
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags = ( Flags & ~UTYPE_VERBOSE_HEADER ) | UTYPE_HEADER;
			else if ( !_wcsicmp( arg[i], L"/HH" ) )
				Flags |= ( UTYPE_HEADER | UTYPE_VERBOSE_HEADER );
			else if ( !_wcsnicmp( arg[i], L"/HW", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, (unsigned int*) &HexWidth, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad HexWidth: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( ( HexWidth < 0 ) || ( HexWidth > 64 ) || ( HexWidth % 8 ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"HexWidth must be 8 to 64 and a multiple of 8: %u\r\n", HexWidth );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/K" ) ) {
				TabWidth = UINT_MAX;
				TabWidthSpecified = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/K", 2 )) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &TabWidth, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad tab size \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( TabWidth > MAX_TAB_SIZE ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Tab size must be 0 - %d\r\n", MAX_TAB_SIZE );
					goto ErrorExit;
				}

				TabWidthSpecified = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/LF" ) )
				Flags = Flags & ~UTYPE_CR_ONLY | UTYPE_LF_ONLY;
			else if ( !_wcsicmp( arg[i], L"/L" ) ) {
				Flags |= UTYPE_LINE_NUMBERS;
				LineNumberFormat = NULL;
			}
			else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				Flags |= UTYPE_LINE_NUMBERS;
				if ( arg[i][j] )
					LineNumberFormat = arg[i] + j;
				else
					LineNumberFormat = NULL;
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j] ) ) {
						case 'B' :
							WriteBOM = FALSE;
							break;

						case 'C' :
							NoHighlight = TRUE;
							break;

						case 'D' :
							Recurse |= FH_FLAG_NO_HIDDEN_DIRS;
							break;

						case 'F' :
							Recurse |= FH_FLAG_SUPPRESS_FNF_ERROR;
							break;

						case 'H' :
							Flags &= ~UTYPE_HANDBRAKE;
							break;

						case 'J' :
							Recurse |= FH_FLAG_NO_JUNCTIONS;
							break;

						case 'Z' :
							Recurse |= FH_FLAG_NO_SYSTEM_DIRS;
							break;

						default :
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c\r\n", towupper( arg[i][j] ) );
							goto ErrorExit;
					}

					j++;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= UTYPE_FIX_QUOTES;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Flags |= UTYPE_TRIM;
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;

				if ( towupper( arg[i][j] ) == 'U' ) {
					j++;
					Flags |= UTYPE_INCLUDE_UNTIL_STRING;
				}

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( ShowUntilString, MAX_ARG_LEN, arg[i] + j );
				_wcslwr_s( ShowUntilString, MAX_ARG_LEN );
			}
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				MaxLength = UINT_MAX;
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &MaxLength, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad length: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( MaxLength && ( ( MaxLength < 16 ) || ( MaxLength > ( MAX_ARG_LEN - 1 ) ) ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Length must be 0 or %u - %u\r\n", 16, MAX_ARG_LEN - 1 );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Flags |= UTYPE_HEX;
			else if ( !_wcsicmp( arg[i], L"/XS" ) )
				Flags |= UTYPE_HEX | UTYPE_HEX_SPACES;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/Z", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				switch ( towupper( arg[i][j] ) ) {

				case 'N' :
				case '\0' :
					Flags &= ~( UTYPE_SKIP_NULLS | UTYPE_NULLS_TO_INVCHAR );
					break;

				case 'I' :
					Flags = ( Flags & ~UTYPE_SKIP_NULLS ) | UTYPE_NULLS_TO_INVCHAR;
					break;

				case 'S' :
					Flags = ( Flags & ~UTYPE_NULLS_TO_INVCHAR ) | UTYPE_SKIP_NULLS;
					break;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption: /Z:%c\r\n", towupper( arg[i][j] ) );
					goto ErrorExit;
				}

			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( Pause && !TabWidthSpecified && ConsoleHandle( STD_OUTPUT_HANDLE ) )
		TabWidth = UINT_MAX;

	if ( TabWidth > MAX_TAB_SIZE )
		TabWidth = EnvironmentNumericValue( L"TABSIZE", 0, MAX_TAB_SIZE );

	if ( TabWidth > MAX_TAB_SIZE )
		TabWidth = 4;

	
	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );

	if ( ConsoleHandle( STD_OUTPUT_HANDLE ) )
		Flags &= ~( UTYPE_CR_ONLY | UTYPE_LF_ONLY );
	else {
		MaxLength = 0;
		Flags &= ~UTYPE_HANDBRAKE;
	}

	if ( MaxLength == UINT_MAX ) {
		MaxLength = ConsoleWidth();
		if ( MaxLength )
			MaxLength--;
	}

	if ( Flags & UTYPE_FIX_QUOTES )
		InitializeFixQuotesVars();


	if ( ( Flags & UTYPE_HEX ) && !HexWidth ) {
		int ConsoleWidth = CL_GetConsoleWidth( 0 );

		if ( ConsoleWidth >= 276 )
			HexWidth = 64;
		else if ( ConsoleWidth >= 210 )
			HexWidth = 48;
		else if ( ConsoleWidth >= 144 )
			HexWidth = 32;
		else
			HexWidth = 16;
	}


	PassArgs.MaxLength			= MaxLength;
	PassArgs.TabWidth			= TabWidth;
	PassArgs.LineNumbersFormat	= LineNumberFormat;
	PassArgs.ShowAfterString	= ShowAfterString;
	PassArgs.ShowUntilString	= ShowUntilString;
	PassArgs.HexWidth			= HexWidth;


	if ( !StdinSpecified ) {
		if ( ( ConsoleHandle( STD_INPUT_HANDLE ) == 0 ) || !FilenameSpecified ) {
			fhrv = TypeFile( (LPTSTR) StdinFilename, Flags, &PassArgs );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &PassArgs, TypeFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	UTYPE_COMMAND_VERSION


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				TypeFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2+

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

	1.0.1.2		2024-01-05
				Minor tweaks to the handling of null characters in hex mode.

*/
