
//			UNICODIFY command:


#ifndef UNICODIFY_COMMAND_VER
#define UNICODIFY_COMMAND_VER		1.0.1.1


#define UNICODIFY_NO_BOM			0x0001
#define UNICODIFY_QUIET				0x0004
#define UNICODIFY_ANYWAY			0x0008
#define UNICODIFY_OVERWRITE_RO		0x0010
#define UNICODIFY_PROMPT			0x0020
#define UNICODIFY_FIX_EOLS			0x0040
#define UNICODIFY_FIX_QUOTES		0x0080




const wchar_t UnicodifyHelpText[] = L"Convert text files to Unicode.\n\
\n\
UNICODIFY /CP:n /L /O /P /Q /S /T /UTF8 /UTF16 filename....\r\n\
\n\
\t/CP:    interpret OEM text as code page n\n\
\t/L      normalize line endings to CR/LF\n\
\t/O      overwrite read-only files\n\
\t/P      prompt for each file\n\
\t/Q      replace ASCII quotes and apostrophes\n\
\t/S      search in subdirectories for matching files\n\
\t/T      quietly\n\
\t/UTF8   write text as UTF-8\n\
\t/UTF16  write text as UTF-16 (default)\n\
\n\
You must give at least one filename.  This command only converts local\n\
files; standard input, internet files, and CLIP: are not supported.\n";



struct _UnicodifyArgs {
	unsigned int OutputFormat;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};




BYTE UTF8Byte( unsigned int Char, int n )
{
	if ( Char > 0x10ffff )
		return 0;


	switch ( n ) {

	case 1 :

		if ( Char < 0x0080 )
			return (BYTE) Char;
		else if ( Char < 0x0800 )
			return ( Char >> 6 ) | 0xc0;
		else if ( Char < 0x10000 )
			return ( Char >> 12 ) | 0xe0;
		else
			return ( Char >> 18 ) | 0xf0;

	case 2 :

		if ( Char < 0x0080 )
			return 0;
		else if ( Char < 0x0800 )
			return ( Char & 0x003f ) | 0x80;
		else if ( Char < 0x10000 )
			return ( ( Char >> 6 ) & 0x003f ) | 0x80;
		else
			return ( ( Char >> 12 ) & 0x003f ) | 0x80;

	case 3 :

		if ( Char < 0x0800 )
			return 0;
		else if ( Char < 0x10000 )
			return ( Char & 0x003f ) | 0x80;
		else
			return ( ( Char >> 6 ) & 0x003f ) | 0x80;

	case 4 :

		if ( Char < 0x10000 )
			return 0;
		else
			return ( Char & 0x003f ) | 0x80;

	default :

		if ( Char < 0x0080 )
			return 1;
		else if ( Char < 0x0800 )
			return 2;
		else if ( Char < 0x10000 )
			return 3;
		else
			return 4;
	}
}



int UnicodifyFlushBuffer( BYTE *Buffer, DWORD BufferSize, DWORD *BufferIndex, HANDLE FileHandle )
{
	if ( !Buffer || !BufferIndex )
		return -1;

	if ( *BufferIndex == 0 )
		return 0;


	DWORD BytesWritten		= 0;


	if ( !WriteFile( FileHandle, Buffer, *BufferIndex, &BytesWritten, NULL ) )
		return -2;

	*BufferIndex = 0;
	return 0;
}



int UnicodifyAddChar( BYTE *Buffer, DWORD BufferSize, DWORD *BufferIndex, HANDLE FileHandle, unsigned int Char, unsigned int OutputFormat )
{
	if ( !Buffer || !BufferIndex )
		return -1;


	int n			= 0;

	wchar_t SurrogatePair[2]	= { 0, 0 };


	if ( OutputFormat == FILE_UTF8 )
		n = UTF8Byte( Char, 0 );
	else if ( ( OutputFormat == FILE_UTF32LE ) || ( OutputFormat == FILE_UTF32BE ) )
		n = 4;
	else if ( Char > 0xffff )
		n = 4;
	else
		n = 2;


	if ( ( *BufferIndex + n ) >= BufferSize )
		if ( UnicodifyFlushBuffer( Buffer, BufferSize, BufferIndex, FileHandle ) < 0 )
			return -1;


	switch ( OutputFormat ) {

	case FILE_UTF8 :

		for ( int i = 1; i <= n; i++ )
			Buffer[(*BufferIndex)++] = UTF8Byte( Char, i );

		break;

	case FILE_UTF16BE :

		if ( n == 4 ) {
			MM_CreateSurrogatePair( Char, SurrogatePair );
			Buffer[*BufferIndex]		= SurrogatePair[0] >> 8;
			Buffer[*BufferIndex + 1]	= SurrogatePair[0] & 0xff;
			Buffer[*BufferIndex + 2]	= SurrogatePair[1] >> 8;
			Buffer[*BufferIndex + 3]	= SurrogatePair[1] & 0xff;
			*BufferIndex += n;
		}
		else {
			Buffer[(*BufferIndex)++] = Char >> 8;
			Buffer[(*BufferIndex)++] = Char & 0xff;
		}

		break;

	case FILE_UTF32LE :

		Buffer[(*BufferIndex)++]	= Char & 0x000000ff;
		Buffer[(*BufferIndex)++]	= ( Char & 0x0000ffff ) >> 8;
		Buffer[(*BufferIndex)++]	= ( Char & 0x00ffffff ) >> 16;
		Buffer[(*BufferIndex)++]	= Char >> 24;
		break;

	case FILE_UTF32BE :

		Buffer[(*BufferIndex)++]	= Char >> 24;
		Buffer[(*BufferIndex)++]	= ( Char & 0x00ffffff ) >> 16;
		Buffer[(*BufferIndex)++]	= ( Char & 0x0000ffff ) >> 8;
		Buffer[(*BufferIndex)++]	= Char & 0x000000ff;
		break;

	default :

		if ( n == 4 ) {
			MM_CreateSurrogatePair( Char, (wchar_t*) ( Buffer + *BufferIndex ) );
			*BufferIndex += n;
		}
		else {
			Buffer[(*BufferIndex)++] = Char & 0xff;
			Buffer[(*BufferIndex)++] = Char >> 8;
		}
	}

	return 0;
}



int MakeBackupFilename( LPCTSTR OldFilename, LPTSTR NewFilename, LPTSTR NewExtension )
{
	if ( !OldFilename || !NewFilename )
		return -1;

	if ( OldFilename[0] == '\0' )
		return -1;

	if ( !QueryIsFile( (LPTSTR) OldFilename ) )
		return ERROR_FILE_NOT_FOUND;



	wchar_t DefaultExt[]					= L"bak";
	wchar_t FilePath[MAX_FILENAME_LEN]		= L"";
	wchar_t BaseName[MAX_FILENAME_LEN]		= L"";
	wchar_t temp[MAX_FILENAME_LEN]			= L"";
	wchar_t index[SHORT_BUF_LEN]			= L"";
	wchar_t *ch								= NULL;
	


	if ( !NewExtension )
		NewExtension = DefaultExt;
	else if ( NewExtension[0] == '\0' )
		NewExtension = DefaultExt;


	FH_PathPart( OldFilename, FilePath );
	FH_FilenamePart( OldFilename, BaseName );

	ch = wcsrchr( BaseName, '.' );
	if ( ch )
		*ch = '\0';


	wcscpy_s( temp, MAX_FILENAME_LEN, FilePath );
	wcscat_s( temp, MAX_FILENAME_LEN, BaseName );

	if ( NewExtension[0] != '.' )
		wcscat_s( temp, MAX_FILENAME_LEN, L"." );

	wcscat_s( temp, MAX_FILENAME_LEN, NewExtension );

	if ( !QueryIsFile( temp ) ) {
		wcscpy_s( NewFilename, MAX_FILENAME_LEN, temp );
		return 0;
	}

	for ( unsigned int i = 0; i < 10000; i++ ) {
		wsprintf( index, L"-%u", i );

		wcscpy_s( temp, MAX_FILENAME_LEN, FilePath );
		wcscat_s( temp, MAX_FILENAME_LEN, BaseName );
		wcscat_s( temp, MAX_FILENAME_LEN, index );

		if ( NewExtension[0] != '.' )
			wcscat_s( temp, MAX_FILENAME_LEN, L"." );

		wcscat_s( temp, MAX_FILENAME_LEN, NewExtension );

		if ( !QueryIsFile( temp ) ) {
			wcscpy_s( NewFilename, MAX_FILENAME_LEN, temp );
			return 0;
		}
	}
	
	return ERROR_TOO_MANY_NAMES;
}



int UnicodifyFile( LPCTSTR Filename, int Flags, void *UnicodifyArgs )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t NewFilename[MAX_FILENAME_LEN]		= L"";
	wchar_t AposWord[SHORT_BUF_LEN]				= L"";
	wchar_t InputEncodingString[64]				= L"";
	wchar_t OutputEncodingString[64]			= L"";

	DWORD OriginalFileAttributes				= 0;
	HANDLE FileHandle							= NULL;

	_MMFile MMFile;
	int ch										= SPACE;
	int lastch									= 0;
	int nextch									= 0;

	_UnicodifyArgs *args						= (_UnicodifyArgs*) UnicodifyArgs;

	const unsigned int OutBufSize				= 8192;
	DWORD OutBufIndex							= 0;
	BYTE OutBuf[OutBufSize];
	unsigned __int64 CharsWritten				= 0;

	int rv										= 0;


	OriginalFileAttributes						= GetFileAttributes( Filename );

	if ( OriginalFileAttributes == INVALID_FILE_ATTRIBUTES )
		return ERROR_FILE_NOT_FOUND;

	if ( ( ( Flags & UNICODIFY_OVERWRITE_RO ) == 0 ) && ( OriginalFileAttributes & FILE_ATTRIBUTE_READONLY ) ) {
		if ( ( Flags & UNICODIFY_QUIET ) == 0 )
			Printf( L"Skipping read-only file: %s\r\n", Filename );

		return 0;
	}

	if ( QueryFileSize( (LPTSTR) Filename, 0 ) == 0 ) {
		if ( ( Flags & UNICODIFY_QUIET ) == 0 )
			Printf( L"Skipping empty file: %s\r\n", Filename );

		return 0;
	}

	if ( Flags & UNICODIFY_PROMPT ) {
		_cwprintf( L"Convert file %s?  [Y/N] ", Filename );

		if ( GetYorN() == 0 ) {
			Printf( L"Skipped.\r\n" );
			return 0;
		}
	}


	MakeBackupFilename( Filename, NewFilename, L"original" );

	if ( !MoveFile( Filename, NewFilename ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t rename file: %s\r\n", Filename );
		return 1;
	}


	rv = OpenMMFile( &MMFile, NewFilename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC | MM_FLAG_HIGH_CHARS, args->CodePage, args->EbcdicCodePage );

	if ( rv ) {
		if ( ( rv == ERROR_EMPTY_FILE ) && ( ( Flags & UNICODIFY_QUIET ) == 0 ) )
			Printf( L"File \"%s\" is empty.\r\n", Filename );

		rv = 1;
		goto UnicodifyError;
	}


	CL_HookBreak();


	MM_GetEncodingString( &MMFile, TRUE, TRUE, InputEncodingString, 64 );
	MM_GetEncodingString( args->OutputFormat, ( Flags & UNICODIFY_NO_BOM ) ? 0 : 2, OutputEncodingString, 64 );
	
	if ( ( MMFile.FileEncoding == args->OutputFormat ) && ( ( Flags & UNICODIFY_ANYWAY ) == 0 ) ) {
		if ( ( Flags & UNICODIFY_QUIET ) == 0 ) {
			MM_GetEncodingString( args->OutputFormat, -1, OutputEncodingString, 64 );
			Printf( L"Already %s; not converting: %s\r\n", OutputEncodingString, Filename );
		}

		rv = 0;
		goto UnicodifyError;
	}


	FileHandle = CreateFile( Filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( ( FileHandle == INVALID_HANDLE_VALUE ) || ( FileHandle == NULL ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error creating output file: %s\r\n", Filename );

		rv = 1;
		goto UnicodifyError;
	}


	OutBufIndex = 0;

	if ( ( Flags & UNICODIFY_NO_BOM ) == 0 )
		UnicodifyAddChar( OutBuf, OutBufSize, &OutBufIndex, FileHandle, BOM, args->OutputFormat );


	while ( MMFile.Index < MMFile.FileSize ) {
		ch = MMGetChar( &MMFile );
		if ( ch < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error reading from file: %s\r\n", NewFilename );
			rv = 1;
			goto UnicodifyError;
		}
		else if ( ch == 0 )
			continue;

		if ( ( Flags & UNICODIFY_FIX_EOLS ) && ( _MMIsEOL( ch ) ) ) {
			if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) )
				MMGetChar( &MMFile );

			if ( UnicodifyAddChar( OutBuf, OutBufSize, &OutBufIndex, FileHandle, CR, args->OutputFormat ) < 0 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error writing to file: %s\r\n", Filename );
				rv = 1;
				goto UnicodifyError;
			}

			ch = LF;
			CharsWritten++;
		}
		else if ( ( ch == 0x0085 ) && ( MMFile.FileEncoding == FILE_EBCDIC ) )
			ch = LF;

		if ( Flags & UNICODIFY_FIX_QUOTES ) {
			nextch = MMPeekAhead( &MMFile, 0 );

			if ( ReplaceDoubles && ( ch == APOSTROPHE ) && ( nextch == APOSTROPHE ) ) {
				if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) ) {
					ch = QUOTE;
					MMGetChar( &MMFile );
					nextch = MMPeekAhead( &MMFile, 0 );
				}
			}

			if ( ReplaceDoubles && ReplaceGraves && ( ch == GRAVE ) && ( nextch == GRAVE ) ) {
				if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) ) {
					ch = LDQUOTE;
					MMGetChar( &MMFile );
					nextch = MMPeekAhead( &MMFile, 0 );
				}
			}

			if ( ReplaceGraves && ( ch == GRAVE ) )
				if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) )
					ch = LSQUOTE;

			if ( ( ch == QUOTE ) || ( ch == FWQUOTE ) ) {
				if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || _MMIsEOL( nextch ) || ( nextch == '\0' ) ) )
					ch = DoublePrime;
				else if ( iswalpha( lastch ) && !iswalpha( nextch ) )
					ch = CloseQuote;
				else if ( iswalpha( nextch ) && !iswalpha( lastch ) )
					ch = OpenQuote;
				else if ( iswgraph( lastch ) && ( lastch != OpenSQuote ) )
					ch = CloseQuote;
				else {
					ch = OpenQuote;

					if ( iswspace( lastch ) )
						if ( _MMIsEOL( nextch ) || ( nextch == '\0' ) )
							ch = CloseQuote;
				}
			}
			else if ( ( ch == APOSTROPHE ) || ( ch == FWAPOSTROPHE ) ) {
				if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
					ch = CloseSQuote;

					if ( iswalnum( lastch ) && iswalpha( nextch ) )
						ch = Apostrophe;
				}
				else {
					ch = OpenSQuote;

					if ( iswspace( lastch ) )
						if ( _MMIsEOL( nextch ) || ( nextch == '\0' ) )
							ch = CloseSQuote;
				}

				if ( !iswalnum( lastch ) && iswalpha( nextch ) ) {
					//		check for word beginning with apostrophe ....

					AposWord[0] = APOSTROPHE;
					int peek = 0;
					for ( int j = 0; j < 15; j++ ) {
						peek = MMPeekAhead( &MMFile, j );
						AposWord[j + 1] = ( peek < SPACE ) ? SPACE : peek;
					}

					AposWord[16] = 0x00;
					if ( IsWordBeginningWithApostrophe( AposWord ) )
						ch = Apostrophe;
				}
				else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
					ch = Okina;
					
				if ( ( iswalnum( lastch ) == 0 ) && iswdigit( nextch ) )
					if ( iswdigit( MMPeekAhead( &MMFile, 1 ) ) )
						if ( iswalnum( MMPeekAhead( &MMFile, 2 ) ) == 0 )
							ch = Apostrophe;

				if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || _MMIsEOL( nextch ) || ( nextch == '\0' ) ) )
					ch = Prime;
			}
			else if ( ( ch == '-' ) || ( ch == 0xff0d ) ) {
				if ( ( ( lastch != '-' ) && ( lastch != 0xff0d ) ) && ( nextch == ch ) ) {
					int tch = MMPeekAhead( &MMFile, 1 );
					if ( ( tch != '-' ) && ( tch != 0xff0d ) ) {
						ch =EmDash;
						MMGetChar( &MMFile );
					}
				}
			}
		}
		
		if ( UnicodifyAddChar( OutBuf, OutBufSize, &OutBufIndex, FileHandle, ch, args->OutputFormat ) < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error writing to file: %s\r\n", Filename );
			rv = 1;
			goto UnicodifyError;
		}

		lastch = ch;
		CharsWritten++;
	}

	if ( UnicodifyFlushBuffer( OutBuf, OutBufSize, &OutBufIndex, FileHandle ) < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error writing to file: %s\r\n", Filename );
		rv = 1;
		goto UnicodifyError;
	}
	

	CloseMMFile( &MMFile );
	CloseHandle( FileHandle );

	SetFileAttributes( Filename, OriginalFileAttributes );

	if ( ( Flags & UNICODIFY_QUIET ) == 0 ) {
		Printf( L"Converted from %s to %s: %s\r\n", InputEncodingString, OutputEncodingString, Filename );
		Printf( L"%I64u character%s.\r\n", CharsWritten, ( CharsWritten == 1 ) ? L"" : L"s" );
	}

	CL_ReleaseBreak();
	return 0;


UnicodifyError:

	CloseMMFile( &MMFile );

	if ( FileHandle )
		CloseHandle( FileHandle );

	if ( NewFilename[0] ) {
		if ( QueryIsFile( NewFilename ) ) {
			DeleteFile( Filename );
			MoveFile( NewFilename, Filename );
		}
	}
	
	CL_ReleaseBreak();
	return rv;
}



DLLExports int WINAPI unicodify( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;

	BOOL FilenameSpecified	= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	_UnicodifyArgs Args		= { 0, 0, 0 };



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( UnicodifyHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( UNICODIFY_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				Flags |= UNICODIFY_FIX_EOLS;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, NULL, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Flags |= UNICODIFY_OVERWRITE_RO;
			else if ( !_wcsicmp( arg[i], L"/P" ) )
				Flags |= UNICODIFY_PROMPT;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= UNICODIFY_FIX_QUOTES;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Flags |= UNICODIFY_QUIET;
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;

				if ( !_wcsnicmp( arg[i] + j, L"TF", 2 ) )
					j += 2;

				if ( arg[i][j] == '-' )
					j++;

				if ( !_wcsicmp( arg[i] + j, L"8" ) )
					Args.OutputFormat = FILE_UTF8;
				else if ( !_wcsicmp( arg[i] + j, L"16" ) )
					Args.OutputFormat = FILE_UTF16LE;
				else if ( !_wcsicmp( arg[i] + j, L"16LE" ) )
					Args.OutputFormat = FILE_UTF16LE;
				else if ( !_wcsicmp( arg[i] + j, L"16BE" ) )
					Args.OutputFormat = FILE_UTF16BE;
				else if ( !_wcsicmp( arg[i] + j, L"32" ) )
					Args.OutputFormat = FILE_UTF32LE;
				else if ( !_wcsicmp( arg[i] + j, L"32LE" ) )
					Args.OutputFormat = FILE_UTF32LE;
				else if ( !_wcsicmp( arg[i] + j, L"32BE" ) )
					Args.OutputFormat = FILE_UTF32BE;
				else {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Unknown suboption: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Flags |= UNICODIFY_ANYWAY;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( QueryIsURL( arg[i] ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Internet files not supported: %s\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( !_wcsicmp( arg[i], ClipFilename ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"CLIP: is not a file\r\n" );
					goto ErrorExit;
				}

				FilenameSpecified = TRUE;
			}
		}
	}


	if ( !WriteBOM )
		Flags |= UNICODIFY_NO_BOM;

	if ( Args.OutputFormat == 0 )
		Args.OutputFormat = FILE_UTF16LE;

	if ( !FilenameSpecified ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		goto ErrorExit;
	}


	if ( Flags & UNICODIFY_FIX_QUOTES )
		InitializeFixQuotesVars();



	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, UnicodifyFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv  = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	UNICODIFY_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				UnicodifyFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2+

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
