
//			XFILTER command:


#ifndef XFILTER_COMMAND_VER
#define XFILTER_COMMAND_VER			1.0.1.1


#define XFILTER_NO_BLANKS			1
#define XFILTER_HEADER				2
#define XFILTER_TRIM				4



const wchar_t XFilterHelpText[] = L"Process lines of a file using variable expansion.\n\
\n\
XFILTER /B /CP:n /F:\"format\" /H /P /S /T filename....\r\n\
\n\
\t/B    discard blank lines\n\
\t/CP:  interpret OEM text as code page n\n\
\t/F:   format string; see below\n\
\t/H    display filenames\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\t/T    trim leading and trailing whitespace\n\
\n\
The format string contains TCC variables and functions, which will be\n\
expanded for each line in the file.  Double all percent signs to prevent\n\
variables from being expanded before the command is executed.  An asterisk \n\
in the format string will be replaced with the line from the file.  The\n\
/F:\"format\" option is required.\n\
\n\
Input filenames may be specified on the command line, or text may be\n\
piped or redirected into XFILTER.\n";



struct _XFilterArgs {
	LPTSTR Format;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



int XFilterFile( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename || !Args )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_XFilterArgs *args							= (_XFilterArgs*) Args;

	wchar_t LineBuffer[MAX_ARG_LEN]				= L"";
	wchar_t OldLineVar[MAX_ARG_LEN]				= L"";
	wchar_t LineVarName[]						= L"_LINE";
	wchar_t OldFilenameVar[MAX_FILENAME_LEN]	= L"";
	wchar_t FilenameVarName[]					= L"_FILENAME";
	wchar_t temp[MAX_ARG_LEN]					= L"";

	_MMFile MMFile;

	unsigned int LineNumber				= 0;
	wchar_t ch							= '\0';
	int j								= 0;

	BOOL SkipBlanks		= ( Flags & XFILTER_NO_BLANKS ) > 0;
	BOOL TrimSpaces		= ( Flags & XFILTER_TRIM ) > 0;

	LPTSTR Format = args->Format;

	if ( !Format )
		return -1;
	else if ( Format[0] == '\0' )
		return 0;

	GetEnvironmentVariable( LineVarName, OldLineVar, MAX_ARG_LEN );
	GetEnvironmentVariable( FilenameVarName, OldFilenameVar, MAX_FILENAME_LEN );


	//  Open file:
	int rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	if ( Flags & XFILTER_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;

	SetEnvironmentVariable( FilenameVarName, Filename );


	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, MAX_ARG_LEN, SkipBlanks ) )
			break;

		wsprintf( temp, L"%u", LineNumber++ );
		SetEnvironmentVariable( LineVarName, temp );

		temp[0] = '\0';
		j = 0;

		if ( TrimSpaces )
			TrimLine( LineBuffer );

		SafeChars( LineBuffer );

		for ( int i = 0; Format[i]; i++ ) {
			ch = Format[i];

			if ( ch == '*' ) {
				wcscat_s( temp, MAX_ARG_LEN, LineBuffer );
				j = (int) wcslen( temp );
			}
			else {
				if ( j < ( MAX_ARG_LEN - 1 ) ) {
					temp[j++] = ch;
					temp[j] = '\0';
				}
			}
		}

		ExpandVariables( temp, 1 );
		UnsafeChars( temp );

		QPuts( temp );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}


	CloseMMFile( &MMFile );
	SetEnvironmentVariable( LineVarName, OldLineVar[0] ? OldLineVar : NULL );
	SetEnvironmentVariable( FilenameVarName, OldFilenameVar[0] ? OldFilenameVar : NULL );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	SetEnvironmentVariable( LineVarName, OldLineVar[0] ? OldLineVar : NULL );
	SetEnvironmentVariable( FilenameVarName, OldFilenameVar[0] ? OldFilenameVar : NULL );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI xfilter( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t FormatString[MAX_ARG_LEN]	= L"";

	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	_XFilterArgs Args		= { NULL, 0, 0 };


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( XFilterHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( XFILTER_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Flags |= XFILTER_NO_BLANKS;
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] )
					wcscpy_s( FormatString, MAX_ARG_LEN, arg[i] + j );
				else {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing format string\r\n" );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= XFILTER_HEADER;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Flags |= XFILTER_TRIM;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}


	if ( FormatString[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing format string\r\n" );
		goto ErrorExit;
	}


	Args.Format = FormatString;


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv =  XFilterFile( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, XFilterFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	XFILTER_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				XFilterFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
